###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import shutil
import tempfile
import unittest

from gtdbtk.config.output import PFAM_TOP_HIT_SUFFIX, TIGRFAM_TOP_HIT_SUFFIX
from gtdbtk.io.marker.tophit import Hit, TopHitPfamFile, TopHitFile, TopHitTigrFile


class TestHit(unittest.TestCase):

    def test___init__(self):
        hit = Hit('gene_id', 'hmm_id', 0.05, 100)
        self.assertEqual('gene_id', hit.gene_id)
        self.assertEqual('hmm_id', hit.hmm_id)
        self.assertAlmostEqual(0.05, hit.e_val, 2)
        self.assertAlmostEqual(100, hit.bit_score, 0)

    def test___repr__(self):
        hit = Hit('a', 'b', 0.05, 100)
        self.assertEqual(f'a b (0.05/100)', repr(hit))

    def test___eq___equal(self):
        self.assertEqual(Hit('gene_id', 'hmm_id', 1e-100, 1e-100),
                         Hit('gene_id', 'hmm_id', 1e-100, 1e-100))

    def test___eq___not_equal(self):
        self.assertNotEqual(Hit('gene_id', 'hmm_id', 0.01, 100),
                            Hit('gene+id', 'hmm_id', 0.01, 100))
        self.assertNotEqual(Hit('gene_id', 'hmm_id', 0.01, 100),
                            Hit('gene_id', 'hmm+id', 0.01, 100))
        self.assertNotEqual(Hit('gene_id', 'hmm_id', 1e-100, 100),
                            Hit('gene_id', 'hmm_id', 1e-101, 100))
        self.assertNotEqual(Hit('gene_id', 'hmm_id', 0.01, 1e-100),
                            Hit('gene_id', 'hmm_id', 0.01, 1e-101))

    def test___lt__(self):
        self.assertLess(Hit('gene_id', 'hmm_id', 1e-100, 10),
                        Hit('gene_id', 'hmm_id', 1e-100, 20))
        self.assertLess(Hit('gene_id', 'hmm_id', 1e-100, 10),
                        Hit('gene_id', 'hmm_id', 1e-200, 10))
        self.assertLess(Hit('gene_id', 'z', 1e-100, 10),
                        Hit('gene_id', 'a', 1e-100, 10))
        self.assertLess(Hit('z', 'hmm_id', 1e-100, 10),
                        Hit('a', 'hmm_id', 1e-100, 10))

    # def test___gt__(self):
    #     self.assertGreater(Hit('gene_id', 'hmm_id', 1e-100, 20),
    #                        Hit('gene_id', 'hmm_id', 1e-100, 10))
    #     self.assertGreater(Hit('gene_id', 'hmm_id', 1e-200, 10),
    #                        Hit('gene_id', 'hmm_id', 1e-100, 10))
    #     self.assertGreater(Hit('gene_id', 'a', 1e-100, 10),
    #                        Hit('gene_id', 'z', 1e-100, 10))
    #     self.assertGreater(Hit('a', 'hmm_id', 1e-100, 10),
    #                        Hit('z', 'hmm_id', 1e-100, 10))

    def test___hash__(self):
        self.assertEqual(hash(Hit('a', 'b', 0.1, 10)),
                         hash(Hit('a', 'b', 0.1, 10)))
        self.assertNotEqual(hash(Hit('a', 'b', 0.1, 10)),
                            hash(Hit('a', 'b', 0.1, 20)))
        self.assertNotEqual(hash(Hit('a', 'b', 0.1, 10)),
                            hash(Hit('a', 'b', 0.2, 10)))
        self.assertNotEqual(hash(Hit('a', 'b', 0.1, 10)),
                            hash(Hit('a', 'c', 0.1, 10)))
        self.assertNotEqual(hash(Hit('a', 'b', 0.1, 10)),
                            hash(Hit('c', 'b', 0.1, 10)))

    def test_hmm_str(self):
        hit = Hit('gene_id', 'hmm_id', 0.05, 100)
        self.assertEqual('hmm_id,0.05,100', hit.hmm_str())

    def test_ordering(self):
        expected = [Hit('a', 'a', 0.05, 50),
                    Hit('a', 'b', 0.99, 50),
                    Hit('a', 'x', 0.01, 10),
                    Hit('a', 'y', 0.01, 10)]
        test = [expected[2], expected[1], expected[0], expected[3]]
        self.assertListEqual(expected, sorted(test, reverse=True))


class TestTopHitFile(unittest.TestCase):

    def setUp(self):
        self.dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        self.path = os.path.join(self.dir_tmp, 'top_hit.tsv')
        self.th = TopHitFile(self.path)

    def tearDown(self):
        self.th = None
        shutil.rmtree(self.dir_tmp)

    def test_add_hit(self):
        self.th.add_hit('a', 'b', 0.1, 10)
        self.th.add_hit('a', 'b', 0.1, 20)
        self.th.add_hit('a', 'c', 0.1, 10)
        self.th.add_hit('x', 'y', 0.1, 10)
        self.th.add_hit('x', 'y', 0.2, 10)
        self.assertEqual(Hit('a', 'b', 0.1, 20), self.th.get_top_hit('a'))
        self.assertEqual(Hit('x', 'y', 0.1, 10), self.th.get_top_hit('x'))

    def test_contains_gene_id(self):
        self.th.add_hit('a', 'b', 0.1, 10)
        self.assertTrue(self.th.contains_gene_id('a'))
        self.assertFalse(self.th.contains_gene_id('b'))

    def test_contains_gene_hmm(self):
        self.th.add_hit('a', 'b', 0.1, 10)
        self.assertTrue(self.th.contains_gene_hmm('a', 'b'))
        self.assertFalse(self.th.contains_gene_hmm('a', 'c'))
        self.assertFalse(self.th.contains_gene_hmm('c', 'b'))
        self.assertFalse(self.th.contains_gene_hmm('x', 'y'))

    def test_get_top_hit(self):
        self.th.add_hit('a', 'b', 0.1, 10)
        self.th.add_hit('a', 'c', 0.1, 20)
        self.th.add_hit('a', 'd', 0.1, 5)
        self.assertEqual(Hit('a', 'c', 0.1, 20), self.th.get_top_hit('a'))
        self.assertIsNone(self.th.get_top_hit('x'))

    def test_get_hmm_hit(self):
        self.th.add_hit('a', 'b', 0.1, 10)
        self.th.add_hit('a', 'c', 0.1, 20)
        self.assertEqual(Hit('a', 'b', 0.1, 10), self.th.get_hmm_hit('a', 'b'))

    def test_write(self):
        self.th.add_hit('a', 'b', 0.1, 10)
        self.th.add_hit('x', 'y', 0.1, 10)
        self.th.add_hit('a', 'c', 0.1, 20)
        self.th.write()

        with open(self.path, 'r') as fh:
            lines = fh.readlines()
        self.assertEqual(3, len(lines))
        self.assertEqual('Gene Id\tTop hits (Family id,e-value,bitscore)\n', lines[0])
        self.assertEqual('a\tc,0.1,20;b,0.1,10\n', lines[1])
        self.assertEqual('x\ty,0.1,10\n', lines[2])

    def test_read(self):
        with open(self.path, 'w') as fh:
            fh.write('Gene Id Top hits (Family id,e-value,bitscore)\n')
            fh.write('a\tc,0.1,20;b,0.1,10\n')
            fh.write('x\ty,0.1,10\n')
        self.th.read()
        self.assertEqual(Hit('a', 'c', 0.1, 20), self.th.get_hmm_hit('a', 'c'))
        self.assertEqual(Hit('a', 'b', 0.1, 10), self.th.get_hmm_hit('a', 'b'))
        self.assertEqual(Hit('x', 'y', 0.1, 10), self.th.get_hmm_hit('x', 'y'))

    def test_iter_hits(self):
        self.th.add_hit('a', 'b', 0.1, 10)
        self.th.add_hit('x', 'y', 0.2, 20)
        expected = {('a', Hit('a', 'b', 0.1, 10)), ('x', Hit('x', 'y', 0.2, 20))}
        self.assertSetEqual(expected, set(self.th.iter_hits()))


class TestTopHitPfamFile(unittest.TestCase):

    def setUp(self):
        self.dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        self.path = os.path.join(self.dir_tmp, 'top_hit.tsv')
        self.th = TopHitPfamFile(self.path, 'g')

    def tearDown(self):
        self.th = None
        shutil.rmtree(self.dir_tmp)

    def test___init__(self):
        self.assertEqual(os.path.join(self.path, 'g', f'g{PFAM_TOP_HIT_SUFFIX}'), self.th.path)


class TestTopHitTigrFile(unittest.TestCase):

    def setUp(self):
        self.dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        self.th = TopHitTigrFile(self.dir_tmp, 'g')

    def tearDown(self):
        self.th = None
        shutil.rmtree(self.dir_tmp)

    def test___init__(self):
        self.assertEqual(os.path.join(self.dir_tmp, 'g', f'g{TIGRFAM_TOP_HIT_SUFFIX}'), self.th.path)

    def test_add_hit(self):
        self.th.add_hit('a', 'b', 0.1, 20)
        self.th.add_hit('a', 'c', 0.1, 10)
        self.th.add_hit('x', 'y', 0.1, 20)
        self.th.add_hit('x', 'z', 0.0, 20)
        expected = {('a', Hit('a', 'b', 0.1, 20)), ('x', Hit('x', 'z', 0.0, 20))}
        self.assertSetEqual(expected, set(self.th.iter_hits()))



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import shutil
import tempfile
import unittest

from gtdbtk.config.output import PATH_TLN_TABLE_SUMMARY
from gtdbtk.exceptions import GTDBTkExit
from gtdbtk.io.prodigal.tln_table_summary import TlnTableSummaryFile


class TestTlnTableSummaryFile(unittest.TestCase):

    def setUp(self):
        self.dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')

    def tearDown(self):
        self.th = None
        shutil.rmtree(self.dir_tmp)

    def test___init__(self):
        tln = TlnTableSummaryFile(self.dir_tmp, 'tst')
        self.assertDictEqual(tln.genomes, {})
        self.assertEqual(os.path.join(self.dir_tmp, PATH_TLN_TABLE_SUMMARY.format(prefix='tst')), tln.path)

    def test_add_genome(self):
        tln = TlnTableSummaryFile(self.dir_tmp, 'tst')
        tln.add_genome('a', 4)
        tln.add_genome('b', 11)
        self.assertDictEqual({'a': 4, 'b': 11}, tln.genomes)

    def test_add_genome_raises_exception(self):
        tln = TlnTableSummaryFile(self.dir_tmp, 'tst')
        tln.add_genome('a', 4)
        self.assertRaises(GTDBTkExit, tln.add_genome, 'a', 11)

    def test_write(self):
        tln = TlnTableSummaryFile(self.dir_tmp, 'tst')
        tln.add_genome('a', 4)
        tln.add_genome('b', 11)
        tln.write()

        lines = set()
        with open(tln.path) as fh:
            [lines.add(x) for x in fh.readlines()]
        self.assertSetEqual({'a\t4\n', 'b\t11\n'}, lines)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################
import argparse
import os
import shutil
import tempfile
import unittest

from gtdbtk.biolib_lite.exceptions import *
from gtdbtk.exceptions import *
from gtdbtk.main import OptionsParser


class TestOptionsParser(unittest.TestCase):

    def setUp(self):
        self.options_parser = OptionsParser('-1')
        self.dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        pass

    def tearDown(self):
        shutil.rmtree(self.dir_tmp)

    def test__verify_genome_id__valid(self):
        """ Test that a valid genome id returns True. """
        self.assertTrue(self.options_parser._verify_genome_id('genome_1'))

    def test__verify_genome_id__invalid(self):
        """ Test that invalid genome ids throw an exception. """
        for c in list('()[],;='):
            self.assertRaises(GTDBTkExit, self.options_parser._verify_genome_id, 'genome%s1' % c)

    def test__genomes_to_process__genome_dir__valid(self):
        """ Test that the expected results are returned when using genome_dir. """
        open(os.path.join(self.dir_tmp, 'genome_1.fna'), 'a').close()
        open(os.path.join(self.dir_tmp, 'genome_2.fna'), 'a').close()
        open(os.path.join(self.dir_tmp, 'other_file.txt'), 'a').close()
        results, tln_table = self.options_parser._genomes_to_process(self.dir_tmp, '', 'fna')
        expected = {'genome_1': os.path.join(self.dir_tmp, 'genome_1.fna'),
                    'genome_2': os.path.join(self.dir_tmp, 'genome_2.fna')}
        self.assertDictEqual(results, expected)

    def test__genomes_to_process__batchfile__valid(self):
        """ Test that the expected results are returned when using batchfile """
        path_batchfile = os.path.join(self.dir_tmp, 'batchfile.txt')
        path_genome_1 = os.path.join(self.dir_tmp, 'genome_1.fna')
        path_genome_2 = os.path.join(self.dir_tmp, 'genome_2.fna')
        open(path_genome_1, 'a').close()
        open(path_genome_2, 'a').close()

        with open(path_batchfile, 'a') as f:
            f.write(f'{path_genome_1}\tgenome_1\n')
            f.write('\n')
            f.write(f'{path_genome_2}\tgenome_2\t4\n')

        results, tln_table = self.options_parser._genomes_to_process('', path_batchfile, 'fna')
        expected = {'genome_1': path_genome_1, 'genome_2': path_genome_2}
        expected_tln = {'genome_2': 4}
        self.assertDictEqual(results, expected)
        self.assertDictEqual(tln_table, expected_tln)

    def test__genomes_to_process__batchfile__invalid_columns(self):
        """ Test that a batchfile containing columns not equal to 2 throws an exception. """
        path_batchfile = os.path.join(self.dir_tmp, 'batchfile.txt')
        path_genome_1 = os.path.join(self.dir_tmp, 'genome_1.fna')
        path_genome_2 = os.path.join(self.dir_tmp, 'genome_2.fna')
        open(path_genome_1, 'a').close()
        open(path_genome_2, 'a').close()

        with open(path_batchfile, 'a') as f:
            f.write('%s\tgenome_1\n' % path_genome_1)
            f.write('\n')
            f.write('%s\tgenome_2\tfoo\n' % path_genome_2)

        self.assertRaises(GTDBTkExit, self.options_parser._genomes_to_process, '', path_batchfile,
                          'fna')

    def test__genomes_to_process__batchfile__blank_genome_path(self):
        """ Test that a batchfile containing a blank genome path throws an exception. """
        path_batchfile = os.path.join(self.dir_tmp, 'batchfile.txt')
        path_genome_1 = os.path.join(self.dir_tmp, 'genome_1.fna')
        path_genome_2 = os.path.join(self.dir_tmp, 'genome_2.fna')
        open(path_genome_1, 'a').close()
        open(path_genome_2, 'a').close()

        with open(path_batchfile, 'a') as f:
            f.write('%s\tgenome_1\n' % path_genome_1)
            f.write('\n')
            f.write('%s\tgenome_2\n' % '')

        self.assertRaises(GTDBTkExit, self.options_parser._genomes_to_process, '', path_batchfile,
                          'fna')

    def test__genomes_to_process__batchfile__blank_genome_id(self):
        """ Test that a batchfile containing a blank genome id throws an exception. """
        path_batchfile = os.path.join(self.dir_tmp, 'batchfile.txt')
        path_genome_1 = os.path.join(self.dir_tmp, 'genome_1.fna')
        path_genome_2 = os.path.join(self.dir_tmp, 'genome_2.fna')
        open(path_genome_1, 'a').close()
        open(path_genome_2, 'a').close()

        with open(path_batchfile, 'a') as f:
            f.write('%s\tgenome_1\n' % path_genome_1)
            f.write('\n')
            f.write('%s\t\n' % path_genome_2)

        self.assertRaises(GTDBTkExit, self.options_parser._genomes_to_process, '', path_batchfile,
                          'fna')

    def test__genomes_to_process__batchfile__duplicate_genome_id(self):
        """ Test that a batchfile containing duplicate genome ids throws an exception. """
        # Branch 1: The number of columns are not equal to 2.
        path_batchfile = os.path.join(self.dir_tmp, 'batchfile.txt')
        path_genome_1 = os.path.join(self.dir_tmp, 'genome_1.fna')
        path_genome_2 = os.path.join(self.dir_tmp, 'genome_2.fna')
        open(path_genome_1, 'a').close()
        open(path_genome_2, 'a').close()

        with open(path_batchfile, 'a') as f:
            f.write('%s\tgenome_1\n' % path_genome_1)
            f.write('\n')
            f.write('%s\tgenome_1\n' % path_genome_2)

        self.assertRaises(GTDBTkExit, self.options_parser._genomes_to_process, '', path_batchfile,
                          'fna')

    # def test__genomes_to_process__batchfile__invalid_genome_id(self):
    #     """ Test that a batchfile containing duplicate genome ids throws an exception. """
    #     # Branch 1: The number of columns are not equal to 2.
    #     path_batchfile_1 = os.path.join(self.dir_tmp, 'batchfile_1.txt')
    #     path_batchfile_2 = os.path.join(self.dir_tmp, 'batchfile_2.txt')
    #     path_batchfile_3 = os.path.join(self.dir_tmp, 'batchfile_3.txt')
    #     path_genome_1 = os.path.join(self.dir_tmp, 'genome_1.fna')
    #     path_genome_2 = os.path.join(self.dir_tmp, 'genome_2.fna')
    #     open(path_genome_1, 'a').close()
    #     open(path_genome_2, 'a').close()
    #
    #     with open(path_batchfile_1, 'a') as f:
    #         f.write('%s\tgenome_1\n' % path_genome_1)
    #         f.write('\n')
    #         f.write('%s\tGB_genome_2\n' % path_genome_2)
    #
    #     with open(path_batchfile_2, 'a') as f:
    #         f.write('%s\tgenome_1\n' % path_genome_1)
    #         f.write('\n')
    #         f.write('%s\tRS_genome_2\n' % path_genome_2)
    #
    #     with open(path_batchfile_3, 'a') as f:
    #         f.write('%s\tgenome_1\n' % path_genome_1)
    #         f.write('\n')
    #         f.write('%s\tUBAgenome_2\n' % path_genome_2)
    #
    #     self.assertRaises(GTDBTkExit, self.options_parser._genomes_to_process, '', path_batchfile_1, 'fna')
    #     self.assertRaises(GTDBTkExit, self.options_parser._genomes_to_process, '', path_batchfile_2, 'fna')
    #     self.assertRaises(GTDBTkExit, self.options_parser._genomes_to_process, '', path_batchfile_3, 'fna')

    def test__genomes_to_process__no_files(self):
        """ Test that an exception is thrown if no files are found to process """
        # Branch 1 : genome_dir is specified
        tmp_genome_dir = tempfile.mkdtemp()
        try:
            self.assertRaises(GTDBTkExit, self.options_parser._genomes_to_process, tmp_genome_dir, '', 'fna')
        finally:
            shutil.rmtree(tmp_genome_dir)

        # Branch 2: batchfile is specified
        tmp_genome_dir = tempfile.mkdtemp()
        try:
            path_batchfile = os.path.join(tmp_genome_dir, 'batchfile.txt')
            open(path_batchfile, 'a').close()
            self.assertRaises(GTDBTkExit, self.options_parser._genomes_to_process, '', path_batchfile, 'fna')
        finally:
            shutil.rmtree(tmp_genome_dir)

    def test_identify__genome_dir_raises_io_exception(self):
        """ Test that the identify method raises an exception on invalid genome_dir """
        options = argparse.ArgumentParser()
        options.genome_dir = os.path.join(tempfile.gettempdir(), 'non-existent-dir')
        self.assertRaises(BioLibDirNotFound, self.options_parser.identify, options)

    def test_identify__batchfile_raises_io_exception(self):
        """ Test that the identify method raises an exception on invalid batchfile """
        options = argparse.ArgumentParser()
        options.genome_dir = None
        options.batchfile = os.path.join(tempfile.gettempdir(), 'non-existent-file.txt')
        self.assertRaises(BioLibFileNotFound, self.options_parser.identify, options)

    def test_align__identify_dir_raises_io_exception(self):
        """ Test that the align method raises an exception on invalid identify dir """
        options = argparse.ArgumentParser()
        options.identify_dir = os.path.join(tempfile.gettempdir(), 'non-existent-dir')
        self.assertRaises(BioLibDirNotFound, self.options_parser.align, options)

    def test_infer__msa_raises_io_exception(self):
        """ Test that the infer method raises an exception on invalid MSA """
        options = argparse.ArgumentParser()
        options.msa_file = os.path.join(tempfile.gettempdir(), 'non-existent-msa.txt')
        self.assertRaises(BioLibFileNotFound, self.options_parser.infer, options)

    def test_run_test(self):
        """Test that the user-test method runs correctly"""
        options = argparse.ArgumentParser()
        options.out_dir = self.dir_tmp
        options.cpus = 3
        self.assertTrue(self.options_parser.run_test(options))

    # def test_run_test__throws_exception(self):
    #     """Test that the user-test method fails correctly"""
    #     options = argparse.ArgumentParser()
    #     options.out_dir = self.dir_tmp
    #     os.mkdir(os.path.join(self.dir_tmp, 'genomes'))
    #     options.cpus = 3
    #     self.assertRaises(GTDBTkTestFailure, self.options_parser.run_test, options)

    def test_classify__align_dir_raises_io_exception(self):
        """ Test that the classify method raises an exception on invalid align dir """
        options = argparse.ArgumentParser()
        options.align_dir = os.path.join(tempfile.gettempdir(), 'non-existent-dir')
        self.assertRaises(BioLibDirNotFound, self.options_parser.classify, options)

    def test_root__no_tree_raises_io_exception(self):
        """ Test that the infer method raises an exception on invalid tree """
        options = argparse.ArgumentParser()
        options.input_tree = os.path.join(tempfile.gettempdir(), 'non-existent-tree.tree')
        self.assertRaises(BioLibFileNotFound, self.options_parser.root, options)

    def test_decorate__no_tree_raises_io_exception(self):
        """ Test that the infer method raises an exception on invalid tree """
        options = argparse.ArgumentParser()
        options.input_tree = os.path.join(tempfile.gettempdir(), 'non-existent-tree.tree')
        self.assertRaises(BioLibFileNotFound, self.options_parser.decorate, options)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = 'Donovan Parks'
__copyright__ = 'Copyright 2014'
__credits__ = ['Donovan Parks']
__license__ = 'GPL3'
__maintainer__ = 'Donovan Parks'
__email__ = 'donovan.parks@gmail.com'

import logging
import ntpath
import os
import shutil
import tempfile
from collections import defaultdict, namedtuple

import numpy as np

from .common import remove_extension, make_sure_path_exists, check_file_exists
from .execute import check_on_path
from .parallel import Parallel
from .seq_io import read_fasta, write_fasta


class Prodigal(object):
    """Wrapper for running Prodigal in parallel."""

    def __init__(self, cpus, verbose=True):
        """Initialization.

        Parameters
        ----------
        cpus : int
            Number of cpus to use.
        verbose : boolean
            Flag indicating if progress should be reported.
        """

        self.logger = logging.getLogger('timestamp')

        check_on_path('prodigal')

        self.cpus = cpus
        self.verbose = verbose

    def _producer(self, genome_file):
        """Apply prodigal to genome with most suitable translation table.

        Parameters
        ----------
        genome_file : str
            Fasta file for genome.
        """

        genome_id = remove_extension(genome_file)

        aa_gene_file = os.path.join(self.output_dir, genome_id + '_genes.faa')
        nt_gene_file = os.path.join(self.output_dir, genome_id + '_genes.fna')
        gff_file = os.path.join(self.output_dir, genome_id + '.gff')

        best_translation_table = -1
        table_coding_density = {4: -1, 11: -1}
        if self.called_genes:
            os.system('cp %s %s' %
                      (os.path.abspath(genome_file), aa_gene_file))
        else:
            seqs = read_fasta(genome_file)

            if len(seqs) == 0:
                self.logger.warning('Cannot call Prodigal on an empty genome. '
                                    'Skipped: {}'.format(genome_file))
                return None

            with tempfile.TemporaryDirectory('gtdbtk_prodigal_tmp_') as tmp_dir:

                # determine number of bases
                total_bases = 0
                for seq in seqs.values():
                    total_bases += len(seq)

                # call genes under different translation tables
                translation_tables = [4, 11]

                for translation_table in translation_tables:
                    os.makedirs(os.path.join(tmp_dir, str(translation_table)))
                    aa_gene_file_tmp = os.path.join(tmp_dir, str(
                        translation_table), genome_id + '_genes.faa')
                    nt_gene_file_tmp = os.path.join(tmp_dir, str(
                        translation_table), genome_id + '_genes.fna')
                    gff_file_tmp = os.path.join(tmp_dir, str(
                        translation_table), genome_id + '.gff')

                    # check if there is sufficient bases to calculate prodigal
                    # parameters
                    if total_bases < 100000 or self.meta:
                        proc_str = 'meta'  # use best pre-calculated parameters
                    else:
                        proc_str = 'single'  # estimate parameters from data

                    # If this is a gzipped genome, re-write the uncompressed genome
                    # file to disk
                    prodigal_input = genome_file
                    if genome_file.endswith('.gz'):
                        prodigal_input = os.path.join(
                            tmp_dir, os.path.basename(genome_file[0:-3]) + '.fna')
                        write_fasta(seqs, prodigal_input)

                    # there may be ^M character in the input file,
                    # the following code is similar to dos2unix command to remove
                    # those special characters.
                    with open(prodigal_input, 'r') as fh:
                        text = fh.read().replace('\r\n', '\n')
                    processed_prodigal_input = os.path.join(
                        tmp_dir, os.path.basename(prodigal_input))
                    with open(processed_prodigal_input, 'w') as fh:
                        fh.write(text)

                    args = '-m'
                    if self.closed_ends:
                        args += ' -c'

                    cmd = 'prodigal %s -p %s -q -f gff -g %d -a %s -d %s -i %s > %s 2> /dev/null' % (args,
                                                                                                     proc_str,
                                                                                                     translation_table,
                                                                                                     aa_gene_file_tmp,
                                                                                                     nt_gene_file_tmp,
                                                                                                     processed_prodigal_input,
                                                                                                     gff_file_tmp)
                    os.system(cmd)

                    # determine coding density
                    prodigalParser = ProdigalGeneFeatureParser(gff_file_tmp)

                    codingBases = 0
                    for seq_id, _seq in seqs.items():
                        codingBases += prodigalParser.coding_bases(seq_id)

                    codingDensity = float(codingBases) / total_bases
                    table_coding_density[translation_table] = codingDensity

                # determine best translation table
                if not self.translation_table:
                    best_translation_table = 11
                    if (table_coding_density[4] - table_coding_density[11] > 0.05) and table_coding_density[4] > 0.7:
                        best_translation_table = 4
                else:
                    best_translation_table = self.translation_table

                shutil.copyfile(os.path.join(tmp_dir, str(best_translation_table),
                                             genome_id + '_genes.faa'), aa_gene_file)
                shutil.copyfile(os.path.join(tmp_dir, str(best_translation_table),
                                             genome_id + '_genes.fna'), nt_gene_file)
                shutil.copyfile(os.path.join(tmp_dir, str(best_translation_table),
                                             genome_id + '.gff'), gff_file)

        return (genome_id, aa_gene_file, nt_gene_file, gff_file, best_translation_table, table_coding_density[4],
                table_coding_density[11])

    def _consumer(self, produced_data, consumer_data):
        """Consume results from producer processes.

         Parameters
        ----------
        produced_data : tuple
            Summary statistics for called genes for a specific genome.
        consumer_data : list
            Summary statistics of called genes for each genome.

        Returns
        -------
        consumer_data: d[genome_id] -> namedtuple(aa_gene_file,
                                                    nt_gene_file,
                                                    gff_file,
                                                    best_translation_table,
                                                    coding_density_4,
                                                    coding_density_11)
            Summary statistics of called genes for each genome.
        """

        ConsumerData = namedtuple(
            'ConsumerData',
            'aa_gene_file nt_gene_file gff_file best_translation_table coding_density_4 coding_density_11')
        if consumer_data is None:
            consumer_data = defaultdict(ConsumerData)

        genome_id, aa_gene_file, nt_gene_file, gff_file, best_translation_table, coding_density_4, coding_density_11 = produced_data

        consumer_data[genome_id] = ConsumerData(aa_gene_file,
                                                nt_gene_file,
                                                gff_file,
                                                best_translation_table,
                                                coding_density_4,
                                                coding_density_11)

        return consumer_data

    def _progress(self, processed_items, total_items):
        """Report progress of consumer processes.

        Parameters
        ----------
        processed_items : int
            Number of genomes processed.
        total_items : int
            Total number of genomes to process.

        Returns
        -------
        str
            String indicating progress of data processing.
        """

        return self.progress_str % (processed_items, total_items, float(processed_items) * 100 / total_items)

    def run(self,
            genome_files,
            output_dir,
            called_genes=False,
            translation_table=None,
            meta=False,
            closed_ends=False):
        """Call genes with Prodigal.

        Call genes with prodigal and store the results in the
        specified output directory. For convenience, the
        called_gene flag can be used to indicate genes have
        previously been called and simply need to be copied
        to the specified output directory.

        Parameters
        ----------
        genome_files : list of str
            Nucleotide fasta files to call genes on.
        called_genes : boolean
            Flag indicating if genes are already called.
        translation_table : int
            Specifies desired translation table, use None to automatically
            select between tables 4 and 11.
        meta : boolean
            Flag indicating if prodigal should call genes with the metagenomics procedure.
        closed_ends : boolean
            If True, do not allow genes to run off edges (throws -c flag).
        output_dir : str
            Directory to store called genes.

        Returns
        -------
        d[genome_id] -> namedtuple(best_translation_table
                                            coding_density_4
                                            coding_density_11)
            Summary statistics of called genes for each genome.
        """

        self.called_genes = called_genes
        self.translation_table = translation_table
        self.meta = meta
        self.closed_ends = closed_ends
        self.output_dir = output_dir

        make_sure_path_exists(self.output_dir)

        progress_func = None
        if self.verbose:
            file_type = 'genomes'
            self.progress_str = '  Finished processing %d of %d (%.2f%%) genomes.'
            if meta:
                file_type = 'scaffolds'
                if len(genome_files):
                    file_type = ntpath.basename(genome_files[0])

                self.progress_str = '  Finished processing %d of %d (%.2f%%) files.'

            self.logger.info('Identifying genes within %s: ' % file_type)
            progress_func = self._progress

        parallel = Parallel(self.cpus)
        summary_stats = parallel.run(
            self._producer, self._consumer, genome_files, progress_func)

        # An error was encountered during Prodigal processing, clean up.
        if not summary_stats:
            shutil.rmtree(self.output_dir)

        return summary_stats


class ProdigalGeneFeatureParser(object):
    """Parses prodigal gene feature files (GFF) output."""

    def __init__(self, filename):
        """Initialization.

        Parameters
        ----------
        filename : str
            GFF file to parse.
        """
        check_file_exists(filename)

        self.genes = {}
        self.last_coding_base = {}

        self.__parseGFF(filename)

        self.coding_base_masks = {}
        for seq_id in self.genes:
            self.coding_base_masks[seq_id] = self.__build_coding_base_mask(
                seq_id)

    def __parseGFF(self, filename):
        """Parse genes from GFF file.

        Parameters
        ----------
        filename : str
            GFF file to parse.
        """
        bGetTranslationTable = True
        with open(filename, 'r') as fh:
            for line in fh:
                if bGetTranslationTable and line.startswith('# Model Data'):
                    data_model_info = line.split(':')[1].strip().split(';')
                    dict_data_model = {}
                    for item in data_model_info:
                        k = item.split('=')[0]
                        v = item.split('=')[1]
                        dict_data_model[k] = v

                    self.translationTable = int(
                        dict_data_model.get('transl_table'))
                    bGetTranslationTable = False

                if line[0] == '#':
                    continue

                line_split = line.split('\t')
                seq_id = line_split[0]
                if seq_id not in self.genes:
                    geneCounter = 0
                    self.genes[seq_id] = {}
                    self.last_coding_base[seq_id] = 0

                geneId = seq_id + '_' + str(geneCounter)
                geneCounter += 1
                start = int(line_split[3])
                end = int(line_split[4])

                self.genes[seq_id][geneId] = [start, end]
                self.last_coding_base[seq_id] = max(
                    self.last_coding_base[seq_id], end)

    def __build_coding_base_mask(self, seq_id):
        """Build mask indicating which bases in a sequences are coding.

        Parameters
        ----------
        seq_id : str
            Unique id of sequence.
        """

        # safe way to calculate coding bases as it accounts
        # for the potential of overlapping genes
        coding_base_mask = np.zeros(self.last_coding_base[seq_id], dtype=np.bool)
        for pos in self.genes[seq_id].values():
            coding_base_mask[pos[0]:pos[1] + 1] = True

        return coding_base_mask

    def coding_bases(self, seq_id, start=0, end=None):
        """Calculate number of coding bases in sequence between [start, end).

        To process the entire sequence set start to 0, and
        end to None.

        Parameters
        ----------
        seq_id : str
            Unique id of sequence.
        start : int
            Start calculation at this position in sequence.
        end : int
            End calculation just before this position in the sequence.
        """

        # check if sequence has any genes
        if seq_id not in self.genes:
            return 0

        # set end to last coding base if not specified
        if end is None:
            end = self.last_coding_base[seq_id]

        return np.sum(self.coding_base_masks[seq_id][start:end])



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import os

from gtdbtk.config.output import PATH_AR122_SUMMARY_OUT, PATH_BAC120_SUMMARY_OUT
from gtdbtk.exceptions import GTDBTkExit


class ClassifySummaryFileRow(object):
    """A row contained within the ClassifySummaryFile object."""

    def __init__(self):
        """Initialise the row, default all of the values to None."""
        self.gid = None
        self.classification = None
        self.fastani_ref = None
        self.fastani_ref_radius = None
        self.fastani_tax = None
        self.fastani_ani = None
        self.fastani_af = None
        self.closest_placement_ref = None
        self.closest_placement_radius = None
        self.closest_placement_tax = None
        self.closest_placement_ani = None
        self.closest_placement_af = None
        self.pplacer_tax = None
        self.classification_method = None
        self.note = None
        self.other_related_refs = None
        self.msa_percent = None
        self.tln_table = None
        self.red_value = None
        self.warnings = None


class ClassifySummaryFile(object):
    """Store the GTDB-Tk classify summary output."""

    def __init__(self, path: str, marker_set: str):
        self.logger = logging.getLogger('timestamp')
        self.path = path
        self.marker_set = marker_set
        self.rows = dict()  # keyed by user_genome
        self.none_value = 'N/A'

    @staticmethod
    def get_col_order(row: ClassifySummaryFileRow = None):
        """Return the column order that will be written. If a row is provided
        then format the row in that specific order."""
        if row is None:
            row = ClassifySummaryFileRow()
        mapping = [('user_genome', row.gid),
                   ('classification', row.classification),
                   ('fastani_reference', row.fastani_ref),
                   ('fastani_reference_radius', row.fastani_ref_radius),
                   ('fastani_taxonomy', row.fastani_tax),
                   ('fastani_ani', row.fastani_ani),
                   ('fastani_af', row.fastani_af),
                   ('closest_placement_reference', row.closest_placement_ref),
                   ('closest_placement_radius', row.closest_placement_radius),
                   ('closest_placement_taxonomy', row.closest_placement_tax),
                   ('closest_placement_ani', row.closest_placement_ani),
                   ('closest_placement_af', row.closest_placement_af),
                   ('pplacer_taxonomy', row.pplacer_tax),
                   ('classification_method', row.classification_method),
                   ('note', row.note),
                   ('other_related_references(genome_id,species_name,radius,ANI,AF)',
                    row.other_related_refs),
                   ('msa_percent', row.msa_percent),
                   ('translation_table', row.tln_table),
                   ('red_value', row.red_value),
                   ('warnings', row.warnings)]
        cols, data = list(), list()
        for col_name, col_val in mapping:
            cols.append(col_name)
            data.append(col_val)
        return cols, data

    def add_row(self, row: ClassifySummaryFileRow):
        if row.gid in self.rows:
            raise GTDBTkExit(f'Attempting to add duplicate row: {row.gid}')
        self.rows[row.gid] = row

    def write(self):
        """Writes the summary file to disk. None will be replaced with N/A"""
        with open(self.path, 'w') as fh:
            fh.write('\t'.join(self.get_col_order()[0]) + '\n')
            for gid, row in sorted(self.rows.items()):
                buf = list()
                for data in self.get_col_order(row)[1]:
                    buf.append(self.none_value if data is None else str(data))
                fh.write('\t'.join(buf) + '\n')

    def read(self):
        """Read the summary file from disk."""
        if not os.path.isfile(self.path):
            raise GTDBTkExit(f'Error, classify summary file not found: {self.path}')
        with open(self.path) as fh:

            # Load and verify the columns match the expected order.
            cols_exp, _ = self.get_col_order()
            cols_cur = fh.readline().strip().split('\t')
            if cols_exp != cols_cur:
                raise GTDBTkExit(f'The classify summary file columns are inconsistent: {cols_cur}')

            # Process the data.
            for line in fh.readlines():
                data = line.strip().split('\t')
                row = ClassifySummaryFileRow()
                row.gid = data[0]
                row.classification = data[1]
                row.fastani_ref = data[2]
                row.fastani_ref_radius = data[3]
                row.fastani_tax = data[4]
                row.fastani_ani = data[5]
                row.fastani_af = data[6]
                row.closest_placement_ref = data[7]
                row.closest_placement_radius = data[8]
                row.closest_placement_tax = data[9]
                row.closest_placement_ani = data[10]
                row.closest_placement_af = data[11]
                row.pplacer_tax = data[12]
                row.classification_method = data[13]
                row.note = data[14]
                row.other_related_refs = data[15]
                row.msa_percent = data[16]
                row.tln_table = data[17]
                row.red_value = data[18]
                row.warnings = data[19]
                self.add_row(row)


class ClassifySummaryFileAR122(ClassifySummaryFile):
    """Store classify summary information for AR122 markers."""

    def __init__(self, out_dir: str, prefix: str):
        path = os.path.join(out_dir, PATH_AR122_SUMMARY_OUT.format(prefix=prefix))
        super().__init__(path, 'ar122')


class ClassifySummaryFileBAC120(ClassifySummaryFile):
    """Store classify summary information for BAC120 markers."""

    def __init__(self, out_dir: str, prefix: str):
        path = os.path.join(out_dir, PATH_BAC120_SUMMARY_OUT.format(prefix=prefix))
        super().__init__(path, 'bac120')



import hashlib
import json
import logging
import math
import os
import random
import re
import time
import urllib.request
from itertools import islice

import dendropy
from tqdm import tqdm

import gtdbtk.config.config as Config
from gtdbtk.config.output import CHECKSUM_SUFFIX
from gtdbtk.exceptions import GTDBTkExit


##################################################
############MISC UTILITIES########################
##################################################


def get_reference_ids():
    """Create a set of reference IDs using the config taxonomy file. This set
    contains the base id and the NCBI formatted ID (e.g. GB_GCA.. and  GCA_...)

    Returns
    -------
    frozenset
        An immutable set with short and long accessions (e.g. GB_GCA_ and GCA_).
    """
    results = set()
    with open(Config.TAXONOMY_FILE) as tf:
        for line in tf:
            raw_id = line.split('\t')[0]
            results.add(raw_id)
            if raw_id[0:4] in ['GCF_', 'GCA_']:
                results.add(add_ncbi_prefix(raw_id))
            elif raw_id[0:3] in ['RS_', 'GB_']:
                results.add(raw_id[3:])
    return frozenset(results)


def add_ncbi_prefix(refname):
    if refname.startswith("GCF_"):
        return "RS_" + refname
    elif refname.startswith("GCA_"):
        return "GB_" + refname
    else:
        return refname


def splitchunks(d, n):
    chunksize = int(math.ceil(len(d) / float(n)))
    it = iter(d)
    for _ in range(0, len(d), chunksize):
        yield {k: d[k] for k in islice(it, chunksize)}


def splitchunks_list(l, n):
    """Yield successive n-sized chunks from l."""
    chunksize = int(math.ceil(len(l) / float(n)))
    for i in range(0, len(l), chunksize):
        yield l[i:i + chunksize]


def generateTempTableName():
    rng = random.SystemRandom()
    suffix = ''
    for _ in range(0, 10):
        suffix += rng.choice(
            'abcefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ')
    return "TEMP" + suffix + str(int(time.time()))


def merge_two_dicts(x, y):
    """Given two dicts, merge them into a new dict as a shallow copy."""
    z = x.copy()
    z.update(y)
    return z


def sha1_dir(path, progress):
    """Recursively add files found within the path and output a SHA1 hash.

    Parameters
    ----------
    path : str
        The path to traverse.
    progress : bool
        True if progress should be displayed to stdout, False otherwise.

    Returns
    -------
    str
        The SHA1 hash representing this directory.
    """
    if progress:
        print('\r[{}]'.format(path), end='', flush=True)

    # Generate a queue of files to process
    queue = list()
    for root, dirs, files in os.walk(path):
        for file in files:
            path_file = os.path.join(root, file)
            queue.append(path_file)
    queue = sorted(queue)

    # Setup the hasher
    block_size = 65536
    hasher = hashlib.sha1()

    # Process the queue and obtain a single hash
    for idx, cur_path in enumerate(queue):

        if progress:
            print('\r[{}] - {}/{} files ({}%)'.format(path,
                                                      idx,
                                                      len(queue),
                                                      round(100 * (idx / len(queue)), 2)),
                  end='', flush=True)

        # Add the hash of the file
        with open(cur_path, 'rb') as fh:
            buf = fh.read(block_size)
            while len(buf) > 0:
                hasher.update(buf)
                buf = fh.read(block_size)

    if progress:
        print('\r' + ' ' * 100, end='', flush=True)
        print('\r', end='', flush=True)

    return hasher.hexdigest()


def sha256(input_file):
    """Determine SHA256 hash for file.

    Parameters
    ----------
    input_file : str
        Name of file.
    Returns
    -------
    str
        SHA256 hash.
    """
    block_size = 65536
    hasher = hashlib.sha1()
    with open(input_file, 'rb') as afile:
        buf = afile.read(block_size)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(block_size)
    return hasher.hexdigest()


def file_has_checksum(file_path, checksum_suffix=CHECKSUM_SUFFIX):
    """Check that the file contents match the checksum.

    Parameters
    ----------
    file_path : str
        Name of the file to check.
    checksum_suffix : str
        Suffix used to denote the file checksum.

    Returns
    -------
    bool
        True if the file has a checksum and it matches the original contents,
        False otherwise.

    """
    check_path = file_path + checksum_suffix
    if os.path.isfile(file_path) and os.path.isfile(check_path):
        with open(check_path, 'r') as check_f:
            return sha256(file_path) == check_f.read()
    return False


def symlink_f(src, dst, force=True):
    """Create a symbolic link pointing to src named dst.

    Parameters
    ----------
    src : str
        The source file.
    dst : str
        The destination file.
    force : bool
        Overwrite any file found with the same name as dst.

    """
    if force and os.path.isfile(dst):
        os.remove(dst)
    os.symlink(src, dst)


def get_memory_gb():
    try:
        with open('/proc/meminfo', 'r') as mem:
            hits = re.findall(r'([\w()]+):\s+(\d+) kB', mem.read())
        return {k: round(int(v) / 1e6, 2) for k, v in hits}
    except Exception:
        return None


def get_proc_memory_gb(pid):
    virt, res = None, None
    try:
        with open(f'/proc/{pid}/status', 'r') as fh:
            contents = fh.read()
        virt = int(re.search(r'VmSize:[^\d]+(\d+)', contents).group(1)) / 1e6
        res = int(re.search(r'VmRSS:[^\d]+(\d+)', contents).group(1)) / 1e6
    finally:
        return virt, res


def get_gtdbtk_latest_version():
    if not Config.GTDBTK_VER_CHECK:
        return None
    try:
        resp = json.loads(urllib.request.urlopen('https://pypi.org/pypi/gtdbtk/json',
                                                 timeout=Config.GTDBTK_VER_TIMEOUT).read().decode('utf-8'))
        return resp['info']['version']
    except Exception:
        return None


class TreeTraversal(object):
    """Efficiently calculates leaf nodes of a given node without re-computing
    any information.
    """

    def __init__(self):
        self.d_node_desc = dict()

    def get_leaf_nodes(self, node):
        """Efficiently return all leaf nodes under a given node.

        Parameters
        ----------
        node : dendropy.Node

        Returns
        -------
        frozenset
            The set of all leaf nodes under the given node.
        """

        # Leaf nodes will always return themselves.
        if node.is_leaf():
            self.d_node_desc[node] = frozenset({node})

        # Stop traversing down if the descendants are already known.
        if node in self.d_node_desc:
            return self.d_node_desc[node]

        # Descendants are not known, traverse down to find them.
        desc_nodes = set()
        for child_node in node.child_node_iter():

            # Already calculated, add it.
            if child_node in self.d_node_desc:
                desc_nodes = desc_nodes.union(self.d_node_desc[child_node])

            # Needs to be calculated, recurse.
            else:
                desc_nodes = desc_nodes.union(self.get_leaf_nodes(child_node))

        # Store the desc and exit.
        self.d_node_desc[node] = frozenset(desc_nodes)
        return self.d_node_desc[node]


def calculate_patristic_distance(qry_node, ref_nodes, tt=None):
    """Computes the patristic distance from the query node to all reference
    nodes. Note that all nodes must be a leaf nodes under max_node.

    Parameters
    ----------
    qry_node : dendropy.Node
        The query taxon node that the distance to all ref nodes will be found.
    ref_nodes : List[dendropy.Node]
        A list of reference nodes that the qry_node will be calculated to.
    tt : Optional[TreeTraversal]
        A TreeTraversal index, if absent a new one will be created.

    Returns
    -------
    Dict[dendropy.Node, float]
        A dictionary keyed by each reference taxon, valued by patristic dist.
    """
    tt = tt or TreeTraversal()

    # Iterate over each of the ref_nodes to find the MRCA to qry_node.
    d_ref_to_mrca = dict()
    for ref_node in ref_nodes:
        cur_dist_to_mrca = ref_node.edge_length

        # Go up the tree until the descendants include qry_node.
        parent_node = ref_node.parent_node
        while parent_node is not None:
            leaf_nodes = tt.get_leaf_nodes(parent_node)

            # Found the MRCA node.
            if qry_node in leaf_nodes:
                d_ref_to_mrca[ref_node] = (parent_node, cur_dist_to_mrca)
                break

            # Keep going up.
            cur_dist_to_mrca += parent_node.edge_length
            parent_node = parent_node.parent_node

        # If the loop did not break, raise an exception.
        else:
            raise GTDBTkExit(f'Unable to find MRCA: {qry_node.taxon.label} / '
                             f'{ref_node.taxon.label}')

    # Compute the distance from the qry_node to each of the MRCAs.
    out = dict()
    for ref_node, (mrca_node, ref_mrca_dist) in d_ref_to_mrca.items():

        # Go up the tree until the MRCA is found again.
        cur_dist_to_mrca = qry_node.edge_length
        cur_node = qry_node.parent_node
        while cur_node is not None:

            # Found the MRCA node.
            if cur_node == mrca_node:
                out[ref_node] = cur_dist_to_mrca + ref_mrca_dist
                break

            # Keep going up.
            cur_dist_to_mrca += cur_node.edge_length
            cur_node = cur_node.parent_node

        # Impossible case, but throw an exception anyway.
        else:
            raise GTDBTkExit(f'Tree is inconsistent: {qry_node.taxon.label} / '
                             f'{ref_node.taxon.label}')

    return out


class tqdm_log(object):
    """A basic wrapper for the tqdm progress bar. Automatically reports the
    runtime statistics after exit.
    """

    def __init__(self, iterable=None, **kwargs):
        # Setup reporting information.
        self.logger = logging.getLogger('timestamp')
        self.start_ts = None

        # Set default parameters.
        default = {'leave': False,
                   'smoothing': 0.1,
                   'bar_format': '==> Processed {n_fmt}/{total_fmt} {unit}s '
                                 '({percentage:.0f}%) |{bar:15}| [{rate_fmt}, ETA {remaining}]'}
        merged = {**default, **kwargs}
        self.args = merged

        # Instantiate tqdm
        self.tqdm = tqdm(iterable, **merged)

    def log(self):
        try:
            # Collect values.
            delta = time.time() - self.start_ts
            n = self.tqdm.n
            unit = self.args.get('unit', 'item')

            # Determine the scale.
            if delta > 60:
                time_unit = 'minute'
                scale = 1 / 60
            elif delta > 60 * 60:
                time_unit = 'hour'
                scale = 1 / (60 * 60)
            elif delta > 60 * 60 * 24:
                time_unit = 'day'
                scale = 1 / (60 * 60 * 24)
            else:
                time_unit = 'second'
                scale = 1

            # Scale units.
            value = delta * scale
            per = n / value

            # Determine what scale to use for the output.
            if per > 1:
                per_msg = f'{per:,.2f} {unit}s/{time_unit}'
            else:
                per_msg = f'{1 / per:,.2f} {time_unit}s/{unit}'

            # Output the message.
            s = 's' if n > 1 else ''
            msg = f'Completed {n:,} {unit}{s} in {value:,.2f} {time_unit}s ({per_msg}).'
            self.logger.info(msg)
        except Exception:
            pass

    def __enter__(self):
        self.start_ts = time.time()
        return self.tqdm

    def __iter__(self):
        try:
            self.start_ts = time.time()
            for item in self.tqdm:
                yield item
            self.log()
        finally:
            self.tqdm.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.log()
        self.tqdm.close()

    def __del__(self):
        self.tqdm.close()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import shutil
import tempfile
import unittest

from gtdbtk.biolib_lite.common import *
from gtdbtk.biolib_lite.exceptions import BioLibFileNotFound, BioLibDirNotFound


class TestBiolibLiteCommon(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_is_float(self):
        """ Check if a string can be converted to a float """
        self.assertTrue(is_float('0.1'))
        self.assertTrue(is_float('-0.1'))
        self.assertTrue(is_float('100.0'))
        self.assertTrue(is_float('1'))
        self.assertFalse(is_float('a'))
        self.assertFalse(is_float(''))

    def test_check_file_exists(self):
        """ Check if a file exists """
        tmp_out_dir = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        try:
            invalid_path = os.path.join(tmp_out_dir, 'foo.txt')
            valid_path = os.path.join(tmp_out_dir, 'test_file.txt')
            open(valid_path, 'w').close()
            self.assertRaises(BioLibFileNotFound, check_file_exists, invalid_path)
            self.assertTrue(check_file_exists(valid_path))
        finally:
            shutil.rmtree(tmp_out_dir)

    def test_check_dir_exists(self):
        """ Check if a directory exists """
        tmp_out_dir = tempfile.mkdtemp()
        try:
            tmp_out_dir = tempfile.mkdtemp()
            self.assertTrue(check_dir_exists(tmp_out_dir))
            self.assertRaises(BioLibDirNotFound, check_dir_exists, '/tmp/path/which/doesnt/exist')
        finally:
            shutil.rmtree(tmp_out_dir)

    def test_make_sure_path_exists(self):
        """ Tests if a path is always created """
        tmp_out_dir = os.path.join(tempfile.gettempdir(), 'tmpgtdb' + next(tempfile._get_candidate_names()))
        try:
            self.assertTrue(make_sure_path_exists(''))
            self.assertTrue(make_sure_path_exists(tmp_out_dir))  # Create the directory.
            self.assertTrue(make_sure_path_exists(tmp_out_dir))  # Return True as it's already created.
            self.assertRaises(BioLibIOException, make_sure_path_exists, '/dev/null/fail')
        finally:
            shutil.rmtree(tmp_out_dir)

    def test_remove_extension(self):
        """ Test that the extension is removed correctly """
        self.assertEqual(remove_extension('foo.root.txt'), 'foo.root')
        self.assertEqual(remove_extension('foo.txt'), 'foo')
        self.assertEqual(remove_extension('foo.root.txt', 'root.txt'), 'foo')
        self.assertEqual(remove_extension('foo'), 'foo')



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = 'Pierre-Alain Chaumeil, Aaron Mussig and Donovan Parks'
__author_email__ = 'p.chaumeil@uq.edu.au'
__copyright__ = 'Copyright 2017'
__credits__ = ['Pierre Chaumeil']
__description__ = 'A toolkit for assigning objective taxonomic classifications to bacterial and archaeal genomes.'
__email__ = 'uqpchaum@uq.edu.au'
__license__ = 'GPL3'
__maintainer__ = 'Pierre-Alain Chaumeil, Aaron Mussig and Donovan Parks'
__maintainer_email__ = 'donovan.parks@gmail.com'
__name__ = 'gtdbtk'
__python_requires__ = '>=3.6'
__status__ = 'Production'
__title__ = 'GTDB-Tk'
__url__ = 'https://github.com/Ecogenomics/GTDBTk'
__version__ = '1.7.0'



import logging
import os
from collections import defaultdict

from gtdbtk.biolib_lite.common import canonical_gid
from gtdbtk.biolib_lite.execute import check_dependencies
from gtdbtk.biolib_lite.taxonomy import Taxonomy
from gtdbtk.config.config import (FASTANI_DIR,
                                  FASTANI_GENOMES_EXT,
                                  FASTANI_GENOME_LIST,
                                  TAXONOMY_FILE,
                                  AF_THRESHOLD)
from gtdbtk.config.output import DIR_ANI_REP_INT_MASH
from gtdbtk.external.fastani import FastANI
from gtdbtk.external.mash import Mash
from gtdbtk.io.gtdb_radii import GTDBRadiiFile


class ANIRep(object):
    """Computes a list of genomes to a list of representatives."""

    def __init__(self, cpus):
        """Instantiate the ANI rep class.

        Parameters
        ----------
        cpus : int
            The maximum number of CPUs available to this workflow.
        """
        self.logger = logging.getLogger('timestamp')
        self.cpus = cpus

    @staticmethod
    def check_dependencies(no_mash):
        """Exits the system if the required programs are not on the path.

        Parameters
        ----------
        no_mash : bool
            True if Mash will be used, False otherwise.
        """
        dependencies = ['fastANI']
        if not no_mash:
            dependencies.append('mash')
        check_dependencies(dependencies)

    @staticmethod
    def _get_ref_genomes():
        """Returns a dictionary of genome accession to genome path.

        Returns
        -------
        dict[str, str]
            Dict[genome_id] = fasta_path
        """
        ref_genomes = dict()
        with open(FASTANI_GENOME_LIST) as g_path_file:
            for line in g_path_file:
                (full_name, path) = line.strip().split()
                if full_name.endswith(FASTANI_GENOMES_EXT):
                    accession = full_name.split(FASTANI_GENOMES_EXT)[0]
                ref_genomes[accession] = os.path.join(FASTANI_DIR, path, full_name)
        return ref_genomes

    def run(self, genomes, no_mash, max_d, out_dir, prefix, mash_k, mash_v, mash_s, min_af, mash_db):
        """Runs the pipeline.

        Parameters
        ----------
        genomes : dict[str, str]
            Dict[genome_id] = fasta_path
        no_mash : bool
            True if Mash should be used for pre-filtering, False otherwise.
        max_d : float
             maximum distance to keep [0-1]
        out_dir : str
            The directory to write the output files to.
        prefix : str
            The prefix to use when writing output files.
        mash_k : int
            k-mer size [1-32]
        mash_v : float
            maximum p-value to keep [0-1]
        mash_s : int
            maximum number of non-redundant hashes
        min_af : float
            alignment fraction to consider closest genome
        mash_db : Optional[str]
            The path to read/write the pre-computed Mash reference sketch database.
        """
        self.check_dependencies(no_mash)

        self.logger.info('Loading reference genomes.')
        ref_genomes = self._get_ref_genomes()
        d_compare = defaultdict(set)
        d_paths = {**genomes, **ref_genomes}

        # Pre-filter using Mash if specified.
        if not no_mash:
            dir_mash = os.path.join(out_dir, DIR_ANI_REP_INT_MASH)

            mash = Mash(self.cpus, dir_mash, prefix)
            self.logger.info(f'Using Mash version {mash.version()}')
            mash_results = mash.run(genomes, ref_genomes, max_d, mash_k, mash_v, mash_s, mash_db)
            for qry_gid, ref_hits in mash_results.items():
                d_compare[qry_gid] = d_compare[qry_gid].union(set(ref_hits.keys()))

        # Compare against all reference genomes.
        else:
            for qry_gid in genomes:
                d_compare[qry_gid] = set(ref_genomes.keys())

        self.logger.info(f'Calculating ANI with FastANI v{FastANI._get_version()}.')
        fastani = FastANI(self.cpus, force_single=True)
        fastani_results = fastani.run(d_compare, d_paths)

        taxonomy = Taxonomy().read(TAXONOMY_FILE, canonical_ids=True)
        ANISummaryFile(out_dir, prefix, fastani_results, taxonomy)
        ANIClosestFile(out_dir,
                       prefix,
                       fastani_results,
                       genomes,
                       min_af,
                       taxonomy)


class ANISummaryFile(object):
    name = 'ani_summary.tsv'

    def __init__(self, root, prefix, results, taxonomy):
        """Writes the ANI summary file generated by this pipeline.

        Parameters
        ----------
        root : str
            The directory to write the summary file to.
        prefix : str
            The output file prefix.
        results: dict[str, dict[str, dict[str, float]]]
            FastANI results.
        taxonomy : dict[str, tuple[str, str, str, str, str, str, str]]
            d[unique_id] -> [d__<taxon>, ..., s__<taxon>]
        """
        self.path = os.path.join(root, f'{prefix}.{self.name}')
        self.results = results
        self.taxonomy = taxonomy
        self.logger = logging.getLogger('timestamp')
        self._write()

    def _write(self):
        with open(self.path, 'w') as fh:
            fh.write('user_genome\treference_genome\tfastani_ani\tfastani_af\treference_taxonomy\n')
            for qry_gid, ref_hits in sorted(self.results.items()):
                for ref_gid, ref_hit in sorted(ref_hits.items(), key=lambda x: (-x[1]['af'], -x[1]['ani'], x[0])):
                    canonical_rid = canonical_gid(ref_gid)
                    taxonomy_str = ';'.join(self.taxonomy[canonical_rid])
                    fh.write(f'{qry_gid}\t{ref_gid}')
                    fh.write(f'\t{ref_hit["ani"]}\t{ref_hit["af"]}')
                    fh.write(f'\t{taxonomy_str}\n')
        self.logger.info(f'Summary of results saved to: {self.path}')


class ANIClosestFile(object):
    name = 'ani_closest.tsv'

    def __init__(self, root, prefix, results, genomes, min_af, taxonomy):
        """Writes the ANI closest file generated by this pipeline.

        Parameters
        ----------
        root : str
            The directory to write the summary file to.
        prefix : str
            The output file prefix.
        results: dict[str, dict[str, dict[str, float]]]
            FastANI results.
        genomes : dict[str, str]
            Dict[genome_id] = fasta_path
        min_af : float
            alignment fraction to consider closest genome
        taxonomy: dict[str, tuple[str, str, str, str, str, str, str]]
            d[unique_id] -> [d__<taxon>, ..., s__<taxon>]
        """
        self.logger = logging.getLogger('timestamp')
        self.path = os.path.join(root, f'{prefix}.{self.name}')
        self.results = results
        self.genomes = genomes
        self.min_af = min_af
        self.taxonomy = taxonomy
        self.gtdb_radii = GTDBRadiiFile()
        self._write()

    def _write(self):
        with open(self.path, 'w') as fh:
            fh.write('user_genome\treference_genome\tfastani_ani\tfastani_af\t'
                     'reference_taxonomy\tsatisfies_gtdb_circumscription_criteria\n')
            for gid in sorted(self.genomes):
                if gid in self.results:
                    thresh_results = [(ref_gid, hit) for (ref_gid, hit) in
                                      self.results[gid].items() if hit['af'] >= self.min_af]
                    closest = sorted(thresh_results, key=lambda x: (-x[1]['ani'], -x[1]['af']))
                    if len(closest) > 0:
                        ref_gid = closest[0][0]
                        canonical_rid = canonical_gid(ref_gid)
                        taxonomy_str = ';'.join(self.taxonomy[canonical_rid])
                        gtdb_ani_radius = self.gtdb_radii.get_rep_ani(canonical_rid)
                        closest_ani = closest[0][1]["ani"]
                        closest_af = closest[0][1]["af"]

                        fh.write(f'{gid}\t{ref_gid}')
                        fh.write(f'\t{closest_ani}\t{closest_af}')
                        fh.write(f'\t{taxonomy_str}')
                        fh.write(f'\t{closest_ani >= gtdb_ani_radius and closest_af >= AF_THRESHOLD}\n')
                    else:
                        fh.write(f'{gid}\tno result\tno result\tno result\tno result\tno result\n')
                else:
                    fh.write(f'{gid}\tno result\tno result\tno result\tno result\n')
        self.logger.info(f'Closest representative hits saved to: {self.path}')



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import sys
import traceback

from gtdbtk import __author__, __copyright__, __version__
from gtdbtk.cli import get_main_parser
from gtdbtk.biolib_lite.exceptions import BioLibError
from gtdbtk.biolib_lite.logger import logger_setup
from gtdbtk.exceptions import *
from gtdbtk.main import OptionsParser
from gtdbtk.tools import get_gtdbtk_latest_version


def print_help():
    print('''\

              ...::: GTDB-Tk v%s :::...

  Workflows:
    classify_wf -> Classify genomes by placement in GTDB reference tree
                     (identify -> align -> classify)
    de_novo_wf  -> Infer de novo tree and decorate with GTDB taxonomy
                     (identify -> align -> infer -> root -> decorate)

  Methods:
    identify -> Identify marker genes in genome
    align    -> Create multiple sequence alignment
    classify -> Determine taxonomic classification of genomes
    infer    -> Infer tree from multiple sequence alignment
    root     -> Root tree using an outgroup
    decorate -> Decorate tree with GTDB taxonomy

  Tools:
    infer_ranks -> Establish taxonomic ranks of internal nodes using RED
    ani_rep     -> Calculates ANI to GTDB representative genomes
    trim_msa    -> Trim an untrimmed MSA file based on a mask
    export_msa  -> Export the untrimmed archaeal or bacterial MSA file

  Testing:
    test          -> Validate the classify_wf pipeline with 3 archaeal genomes 
    check_install -> Verify third party programs and GTDB reference package.

  Use: gtdbtk <command> -h for command specific help
    ''' % __version__)


def main():
    # -------------------------------------------------
    # get and check options
    args = None
    if len(sys.argv) == 1:
        print_help()
        sys.exit(0)
    elif sys.argv[1] in {'-v', '--v', '-version', '--version'}:
        print(f"gtdbtk: version {__version__} {__copyright__} {__author__}")

        # Warn the user they are not using the latest version (if possible)
        latest_ver = get_gtdbtk_latest_version()
        if latest_ver and latest_ver != __version__:
            print(f'There is a newer version of GTDB-Tk available: v{latest_ver}')
        sys.exit(0)
    elif sys.argv[1] in {'-h', '--h', '-help', '--help'}:
        print_help()
        sys.exit(0)
    else:
        args = get_main_parser().parse_args()

    # setup logger
    logger_setup(args.out_dir if hasattr(args, 'out_dir') and args.out_dir else None,
                 "gtdbtk.log", "GTDB-Tk", __version__, False,
                 hasattr(args, 'debug') and args.debug)
    logger = logging.getLogger('timestamp')

    # -------------------------------------------------
    # do what we came here to do
    try:
        gt_parser = OptionsParser(__version__)
        gt_parser.parse_options(args)
    except SystemExit:
        logger.error('Controlled exit resulting from early termination.')
        sys.exit(1)
    except KeyboardInterrupt:
        logger.error('Controlled exit resulting from interrupt signal.')
        sys.exit(1)
    except GTDBTkExit as e:
        if len(str(e)) > 0:
            logger.error('{}'.format(e))
        logger.error('Controlled exit resulting from an unrecoverable error or warning.')
        sys.exit(1)
    except (GTDBTkException, BioLibError) as e:
        msg = 'Controlled exit resulting from an unrecoverable error or warning.\n\n'
        msg += '=' * 80 + '\n'
        msg += 'EXCEPTION: {}\n'.format(type(e).__name__)
        msg += '  MESSAGE: {}\n'.format(e)
        msg += '_' * 80 + '\n\n'
        msg += traceback.format_exc()
        msg += '=' * 80
        logger.error(msg)
        sys.exit(1)
    except Exception as e:
        msg = 'Uncontrolled exit resulting from an unexpected error.\n\n'
        msg += '=' * 80 + '\n'
        msg += 'EXCEPTION: {}\n'.format(type(e).__name__)
        msg += '  MESSAGE: {}\n'.format(e)
        msg += '_' * 80 + '\n\n'
        msg += traceback.format_exc()
        msg += '=' * 80
        logger.error(msg)
        sys.exit(1)


if __name__ == '__main__':
    main()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import multiprocessing as mp
import os
import shutil
import subprocess

from gtdbtk.biolib_lite.prodigal_biolib import Prodigal as BioLibProdigal
from gtdbtk.config.output import CHECKSUM_SUFFIX
from gtdbtk.exceptions import ProdigalException
from gtdbtk.io.prodigal.tln_table import TlnTableFile
from gtdbtk.tools import sha256, file_has_checksum, tqdm_log


class Prodigal(object):
    """Perform ab initio gene prediction using Prodigal."""

    def __init__(self,
                 threads,
                 failed_genomes_file,
                 marker_gene_dir,
                 protein_file_suffix,
                 nt_gene_file_suffix,
                 gff_file_suffix,
                 force):
        """Initialize."""

        self.logger = logging.getLogger('timestamp')
        self.warnings = logging.getLogger('warnings')
        self.failed_genomes_file = failed_genomes_file

        self.threads = threads

        self.marker_gene_dir = marker_gene_dir
        self.protein_file_suffix = protein_file_suffix
        self.nt_gene_file_suffix = nt_gene_file_suffix
        self.gff_file_suffix = gff_file_suffix
        self.force = force
        self.version = self._get_version()

    def _get_version(self):
        try:
            env = os.environ.copy()
            proc = subprocess.Popen(['prodigal', '-v'], stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE, env=env, encoding='utf-8')

            output, error = proc.communicate()

            return error.split('\n')[1].split()[1].replace(':', '')
        except:
            return "(version unavailable)"

    def _run_prodigal(self, genome_id, fasta_path, usr_tln_table):
        """Run Prodigal.

        Parameters
        ----------
        fasta_path : str
            Path to FASTA file to process.
        usr_tln_table : int
            User-specified translation table, None if automatic.
        :return
            False if an error occurred.
        """

        # Create objects to write the output data to.
        output_dir = os.path.join(self.marker_gene_dir, genome_id)
        tln_table_file = TlnTableFile(output_dir, genome_id)

        # Set the paths for output files.
        aa_gene_file = os.path.join(output_dir, genome_id + self.protein_file_suffix)
        nt_gene_file = os.path.join(output_dir, genome_id + self.nt_gene_file_suffix)
        gff_file = os.path.join(output_dir, genome_id + self.gff_file_suffix)
        out_files = (nt_gene_file, gff_file, tln_table_file.path, aa_gene_file)

        # Check if this genome has already been processed (skip).
        if all([file_has_checksum(x) for x in out_files]):
            tln_table_file.read()
            self.warnings.info(f'Skipped Prodigal processing for: {genome_id}')
            return aa_gene_file, nt_gene_file, gff_file, tln_table_file.path, tln_table_file.best_tln_table, True

        # Run Prodigal
        prodigal = BioLibProdigal(1, False)
        summary_stats = prodigal.run([fasta_path], output_dir,
                                     called_genes=False,
                                     translation_table=usr_tln_table)

        # An error occurred in BioLib Prodigal.
        if not summary_stats:
            if self.force:
                return None
            else:
                raise Exception("An error was encountered while running Prodigal.")

        summary_stats = summary_stats[list(summary_stats.keys())[0]]

        # rename output files to adhere to GTDB conventions and desired genome
        # ID
        shutil.move(summary_stats.aa_gene_file, aa_gene_file)
        shutil.move(summary_stats.nt_gene_file, nt_gene_file)
        shutil.move(summary_stats.gff_file, gff_file)

        # save translation table information
        tln_table_file.best_tln_table = summary_stats.best_translation_table
        tln_table_file.coding_density_4 = round(summary_stats.coding_density_4 * 100, 2)
        tln_table_file.coding_density_11 = round(summary_stats.coding_density_11 * 100, 2)
        tln_table_file.write()

        # Create a hash of each file
        for out_file in out_files:
            if out_file is not None:
                with open(out_file + CHECKSUM_SUFFIX, 'w') as fh:
                    fh.write(sha256(out_file))

        return aa_gene_file, nt_gene_file, gff_file, tln_table_file.path, summary_stats.best_translation_table, False

    def _worker(self, out_dict, worker_queue, writer_queue, n_skipped):
        """This worker function is invoked in a process."""

        while True:
            data = worker_queue.get(block=True, timeout=None)
            if data is None:
                break

            genome_id, file_path, usr_tln_table = data

            rtn_files = self._run_prodigal(genome_id, file_path, usr_tln_table)

            # Only proceed if an error didn't occur in BioLib Prodigal
            if rtn_files:
                aa_gene_file, nt_gene_file, gff_file, translation_table_file, best_translation_table, was_skipped = rtn_files
                if was_skipped:
                    with n_skipped.get_lock():
                        n_skipped.value += 1
                prodigal_infos = {"aa_gene_path": aa_gene_file,
                                  "nt_gene_path": nt_gene_file,
                                  "gff_path": gff_file,
                                  "translation_table_path": translation_table_file,
                                  "best_translation_table": best_translation_table}

                out_dict[genome_id] = prodigal_infos
            writer_queue.put(genome_id)

    def _writer(self, num_items, writer_queue):
        """Store or write results of worker threads in a single thread."""
        with tqdm_log(total=num_items,  unit='genome') as p_bar:
            for _ in iter(writer_queue.get, None):
                p_bar.update()

    def run(self, genomic_files, tln_tables):
        """Run Prodigal across a set of genomes.

        Parameters
        ----------
        genomic_files : dict
            Dictionary indicating the genomic and gene file for each genome.
        tln_tables : Dict[str, int]
            Mapping of genome id to user-specified translation table.
        """

        # populate worker queue with data to process
        worker_queue = mp.Queue()
        writer_queue = mp.Queue()

        for genome_id, file_path in genomic_files.items():
            worker_queue.put((genome_id, file_path, tln_tables.get(genome_id)))

        for _ in range(self.threads):
            worker_queue.put(None)

        worker_proc = []
        writer_proc = None
        try:
            manager = mp.Manager()
            out_dict = manager.dict()
            n_skipped = mp.Value('i', 0)

            worker_proc = [mp.Process(target=self._worker, args=(out_dict,
                                                                 worker_queue,
                                                                 writer_queue,
                                                                 n_skipped))
                           for _ in range(self.threads)]
            writer_proc = mp.Process(target=self._writer, args=(len(genomic_files),
                                                                writer_queue))

            writer_proc.start()
            for p in worker_proc:
                p.start()

            for p in worker_proc:
                p.join()

                # Gracefully terminate the program.
                if p.exitcode != 0:
                    raise ProdigalException('Prodigal returned a non-zero exit code.')

            writer_queue.put(None)
            writer_proc.join()
        except Exception as e:
            for p in worker_proc:
                p.terminate()
            if writer_proc:
                writer_proc.terminate()
            raise ProdigalException(f'An exception was caught while running Prodigal: {e}')

        # Report if any genomes were skipped due to having already been processed.
        if n_skipped.value > 0:
            genome_s = 'genome' if n_skipped.value == 1 else 'genomes'
            self.logger.warning(f'Prodigal skipped {n_skipped.value} {genome_s} '
                                f'due to pre-existing data, see warnings.log')

        # Report on any genomes which failed to have any genes called
        result_dict = dict()
        lq_gids = list()
        fails = open(self.failed_genomes_file,'w')
        for gid, gid_dict in out_dict.items():
            if os.path.getsize(gid_dict['aa_gene_path']) <= 1:
                lq_gids.append(gid)
            else:
                result_dict[gid] = gid_dict

        if len(lq_gids) > 0:
            self.logger.warning(f'Skipping {len(lq_gids)} of {len(genomic_files)} '
                                f'genomes as no genes were called by Prodigal. '
                                f'Check the genome quality (see gtdbtk.warnings.log).')
            self.warnings.warning(f'The following {len(lq_gids)} genomes have '
                                  f'been excluded from analysis due to Prodigal '
                                  f'failing to call any genes:')


            # If there are few low-quality genomes just output to console.
            if len(lq_gids) > 10:
                for lq_gid in lq_gids:
                    self.warnings.info(lq_gid)
                    fails.write(f'{lq_gid}\tno genes were called by Prodigal\n')
            else:
                for lq_gid in lq_gids:
                    self.logger.warning(f'Skipping: {lq_gid}')
                    self.warnings.info(lq_gid)
                    fails.write(f'{lq_gid}\tno genes were called by Prodigal\n')

        fails.close()
        return result_dict



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import os

import gtdbtk.config.config as Config
from gtdbtk.biolib_lite.execute import check_dependencies
from gtdbtk.biolib_lite.logger import colour
from gtdbtk.biolib_lite.seq_io import read_fasta
from gtdbtk.exceptions import GTDBTkException, GTDBTkExit
from gtdbtk.tools import sha1_dir


class Misc(object):

    def __init__(self):
        """Initialize."""
        self.logger = logging.getLogger('timestamp')

    def trim_msa(self, untrimmed_msa, mask_type, maskid, output_file):
        """Trim the multiple sequence alignment using a mask.

        Parameters
        ----------
        untrimmed_msa : str
            The path to the untrimmed MSA.
        mask_type : str
            Which mask should be used, reference or user specified.
        maskid : str
            The path to the mask used for trimming.
        output_file : str
            The path to the output trimmed MSA.
        """
        if maskid == 'bac' and mask_type == 'reference':
            mask = os.path.join(Config.MASK_DIR, Config.MASK_BAC120)
        elif maskid == 'arc' and mask_type == 'reference':
            mask = os.path.join(Config.MASK_DIR, Config.MASK_AR122)
        elif mask_type == 'file':
            mask = maskid
        else:
            self.logger.error('Command not understood.')
            raise GTDBTkException('Command not understood.')

        with open(mask, 'r') as f:
            maskstr = f.readline()

        with open(output_file, 'w') as outfwriter:
            dict_genomes = read_fasta(untrimmed_msa, False)

            for k, v in dict_genomes.items():
                aligned_seq = ''.join([v[i] for i in range(0, len(maskstr)) if maskstr[i] == '1'])
                fasta_outstr = ">%s\n%s\n" % (k, aligned_seq)
                outfwriter.write(fasta_outstr)

    def checkfile(self, file_path, file_name):
        """Check that a file exists, output the result to the logger.

        Returns
        -------
        bool
            True if the folder exists, False otherwise.
        """
        if os.path.exists(file_path) and os.path.getsize(file_path) > 0:
            self.logger.info("Check file {} ({}): {}".format(
                file_name, file_path, colour('OK', ['bright'], fg='green')))
            return True
        else:
            self.logger.warning("Check file {} ({}): {}".format(
                file_name, file_path, colour('MISSING', ['bright'], fg='red')))
            return False

    def checkfolder(self, folder_path, folder_name):
        """Check that a folder exists, output the result to the logger.

        Returns
        -------
        bool
            True if the folder exists, False otherwise.
        """
        if os.path.isdir(folder_path) and len(os.listdir(folder_path)) > 0:
            self.logger.info("Check folder {} ({}): {}".format(
                folder_name, folder_path, colour('OK', ['bright'], fg='green')))
            return True
        else:
            self.logger.warning("Check folder {} ({}): {}".format(
                folder_name, folder_path, colour('MISSING', ['bright'], fg='red')))
            return False

    def check_install(self):
        """Check that all reference files exist.

        Returns
        -------
        bool
            True if the installation is complete, False otherwise.
        """

        # Check that all programs are on the system path.
        self.logger.info(f'Checking that all third-party software are on the system path:')
        names = {'prodigal', 'hmmsearch', 'fastANI', 'mash', 'pplacer', 'guppy',
                 'FastTree', 'FastTreeMP', 'hmmalign'}
        for name in sorted(names):
            on_path = False
            try:
                on_path = on_path or check_dependencies([name], exit_on_fail=False)
            except:
                pass
            if on_path:
                self.logger.info("         |-- {:16} {}".format(
                    name, colour('OK', ['bright'], fg='green')))
            else:
                self.logger.info("         |-- {:16} {}".format(
                    name, colour('NOT FOUND', ['bright'], fg='yellow')))

        # Assume this was successful unless otherwise observed.
        ok = True

        # Compute the hash for each directory
        self.logger.info(f'Checking integrity of reference package: {Config.GENERIC_PATH}')
        for obj_path, expected_hash in Config.REF_HASHES.items():
            base_name = obj_path[:-1] if obj_path.endswith('/') else obj_path
            base_name = base_name.split('/')[-1]
            user_hash = sha1_dir(obj_path, progress=True)

            if user_hash != expected_hash:
                self.logger.info("         |-- {:16} {}".format(
                    base_name, colour(f'HASH MISMATCH {user_hash}', ['bright'], fg='yellow')))
                ok = False
            else:
                self.logger.info("         |-- {:16} {}".format(
                    base_name, colour('OK', ['bright'], fg='green')))

        if not ok:
            raise GTDBTkExit('Unexpected files were seen, or the reference package is corrupt.')



import os
from shutil import copyfile

from gtdbtk.biolib_lite.common import make_sure_path_exists
from gtdbtk.config.config import CONCAT_AR122, CONCAT_BAC120
from gtdbtk.exceptions import GTDBTkExit
from gtdbtk.model.enum import Domain


def export_msa(domain: Domain, output_file: str):
    """Exports the GTDB MSA to the specified path.

    :param domain: The domain used to determine the marker set.
    :param output_file: The path to write the MSA.
    """
    if domain is Domain.ARCHAEA:
        file_to_export = CONCAT_AR122
    elif domain is Domain.BACTERIA:
        file_to_export = CONCAT_BAC120
    else:
        raise GTDBTkExit(f'Unknown domain: "{domain}"')

    make_sure_path_exists(os.path.dirname(output_file))
    copyfile(file_to_export, output_file)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
from typing import Dict

from gtdbtk.biolib_lite.common import make_sure_path_exists
from gtdbtk.config.output import PATH_TLN_TABLE_SUMMARY
from gtdbtk.exceptions import GTDBTkExit


class TlnTableSummaryFile(object):
    """Records the translation table for one or more genomes."""
    __slots__ = ('path', 'genomes')

    def __init__(self, out_dir: str, prefix: str):
        """Configure paths and initialise storage dictionary."""
        self.path: str = os.path.join(out_dir, PATH_TLN_TABLE_SUMMARY.format(prefix=prefix))
        self.genomes: Dict[str, int] = dict()

    def add_genome(self, genome_id: str, tln_table: int):
        """Record a translation table for a genome."""
        if genome_id in self.genomes:
            raise GTDBTkExit(f'Genome already exists in summary file: {genome_id}')
        self.genomes[genome_id] = tln_table

    def write(self):
        """Write the translation table summary file to disk."""
        make_sure_path_exists(os.path.dirname(self.path))
        with open(self.path, 'w') as fh:
            for genome_id, tln_table in sorted(self.genomes.items()):
                fh.write(f'{genome_id}\t{tln_table}\n')

    def read(self):
        """Read the translation table summary file from disk."""
        if len(self.genomes) > 0:
            raise GTDBTkExit(f'Warning! Attempting to override in-memory values '
                             f'for translation table summary file: {self.path}')
        with open(self.path, 'r') as fh:
            for line in fh.readlines():
                gid, tbl = line.strip().split('\t')
                self.genomes[gid] = int(tbl)



import logging
import os
import re
import subprocess
import tempfile
from collections import defaultdict
from typing import Tuple, Dict
from gtdbtk.biolib_lite.common import make_sure_path_exists
from gtdbtk.exceptions import GTDBTkExit
from gtdbtk.tools import tqdm_log


class Mash(object):
    """Runs Mash against genomes."""

    def __init__(self, cpus, out_dir, prefix):
        """Instantiate the Mash class.

        Parameters
        ----------
        cpus : int
            The maximum number of CPUs available to Mash.
        out_dir : str
            The directory to write output files to.
        prefix : str
            The prefix for all output files.
        """
        self.logger = logging.getLogger('timestamp')
        self.cpus = max(cpus, 1)
        self.out_dir = out_dir
        self.prefix = prefix

    @staticmethod
    def version():
        """Returns the version of mash, or 'unknown' if not known."""
        try:
            proc = subprocess.Popen(['mash', '--version'], stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE, encoding='utf-8')
            stdout, stderr = proc.communicate()
            if len(stdout) > 0:
                return stdout.strip()
            else:
                return 'unknown'
        except Exception:
            return 'unknown'

    def run(self, qry, ref, mash_d, mash_k, mash_v, mash_s, mash_db) -> Dict[str, Dict[str, Tuple[float, float, int, int]]]:
        """Run Mash on a set of reference and query genomes.

        Parameters
        ----------
        qry : dict[str, str]
            The set of query genomes and their path.
        ref : dict[str, str]
            The set of reference genomes and their path.
        mash_d : float
            The maximum distance to report.
        mash_k : int
            The number of k-mers to store for each sequence.
        mash_v : float
            Maximum p-value to report.
        mash_s: int
            Maximum number of non-redundant hashes.
        mash_db : Optional[str]
            The path to read/write the pre-computed Mash reference sketch database.

        Returns
        -------
        dict[query_id][ref_id] = (dist, p_val, shared_numerator, shared_denominator)
        """
        qry_sketch = QrySketchFile(qry, self.out_dir, self.prefix, self.cpus, mash_k, mash_s)
        ref_sketch = RefSketchFile(ref, self.out_dir, self.prefix, self.cpus, mash_k, mash_s, mash_db)

        # Generate an output file comparing the distances between these genomes.
        mash_dists = DistanceFile(qry_sketch, ref_sketch, self.out_dir, self.prefix,
                                  self.cpus, max_d=mash_d, mash_v=mash_v)
        results = mash_dists.read()

        # Convert the results back to the accession
        path_to_qry = {v: k for (k, v) in qry.items()}
        path_to_ref = {v: k for (k, v) in ref.items()}
        out = defaultdict(dict)
        for qry_path, ref_hits in results.items():
            for ref_path, hit in ref_hits.items():
                out[path_to_qry[qry_path]][path_to_ref[ref_path]] = hit
        return out


class DistanceFile(object):
    """The resulting distance file from the mash dist command."""
    name = 'mash_distances.tsv'

    def __init__(self, qry_sketch, ref_sketch, root, prefix, cpus, max_d, mash_v):
        """Create a new Mash distance file using these arguments.

        Parameters
        ----------
        qry_sketch : QrySketchFile
            The query sketch file generated by Mash.
        ref_sketch : RefSketchFile
            The reference sketch file generated by Mash.
        root : str
            The directory where the distance file will be contained.
        prefix : str
            The prefix to use in for distance file.
        cpus : int
            The maximum number of CPUs available to Mash.
        max_d : float
            The maximum distance to consider.
        mash_v : float
            The maximum value to consider.
        """
        self.logger = logging.getLogger('timestamp')
        self.qry_sketch = qry_sketch
        self.ref_sketch = ref_sketch
        self.path = os.path.join(root, f'{prefix}.{self.name}')
        self.cpus = cpus
        self.max_d = max_d
        self.mash_v = mash_v

        self._calculate()

    def _calculate(self):
        self.logger.info('Calculating Mash distances.')
        args = ['mash', 'dist', '-p', self.cpus, '-d', self.max_d, '-v',
                self.mash_v, self.ref_sketch.path, self.qry_sketch.path]
        args = list(map(str, args))
        with open(self.path, 'w') as f_out:
            proc = subprocess.Popen(args, stdout=f_out,
                                    stderr=subprocess.PIPE, encoding='utf-8')
            _, stderr = proc.communicate()
        if proc.returncode != 0:
            raise GTDBTkExit(f'Error running Mash dist: {proc.stderr.read()}')

    def read(self) -> Dict[str, Dict[str, Tuple[float, float, int, int]]]:
        """Reads the results of the distance file.

        Returns
        -------
        dict[query_id][ref_id] = (dist, p_val, shared_numerator, shared_denominator)
        """
        out = defaultdict(dict)
        with open(self.path, 'r') as fh:
            hits = re.findall(r'(.+)\t(.+)\t(.+)\t(.+)\t(\d+)\/(\d+)\n', fh.read())
        for ref_id, qry_id, dist, p_val, shared_n, shared_d in hits:
            dist, p_val = float(dist), float(p_val)
            shared_num, shared_den = int(shared_n), int(shared_d)
            out[qry_id][ref_id] = (dist, p_val, shared_num, shared_den)
        return out


class SketchFile(object):
    """Output files which are generated by mash sketch."""

    def __init__(self, genomes, path, cpus, k, s):
        """Create a sketch file for a given set of genomes.

        Parameters
        ----------
        genomes : dict[str, str]
            The genomes to create a sketch file from (genome_id, fasta_path).
        path : str
            The path to write the sketch file to.
        cpus : int
            The maximum number of CPUs available for Mash.
        k : int
            The k-mer size.
        s : int
            Maximum number of non-redundant hashes.
        """
        self.logger = logging.getLogger('timestamp')
        self.genomes = genomes
        self.path = path
        self.data = dict()
        self.args = dict()
        self.cpus = cpus
        self.k = k
        self.s = s

        make_sure_path_exists(os.path.dirname(self.path))

        # Use the pre-existing sketch file, otherwise generate it.
        if os.path.isfile(self.path):
            self.logger.info(f'Loading data from existing Mash sketch file: {self.path}')
            self._load_metadata()
            if not self._is_consistent():
                raise GTDBTkExit(f'The sketch file is not consistent with the '
                                 f'input genomes. Remove the existing sketch '
                                 f'file or specify a new output directory.')
        else:
            self.logger.info(f'Creating Mash sketch file: {self.path}')
            self._generate()

    def _load_metadata(self):
        """Loads the metadata from an existing Mash sketch file."""
        args = ['mash', 'info', '-t', self.path]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()

        if proc.returncode != 0:
            raise GTDBTkExit(f'Error reading Mash sketch file {self.path}:\n{stderr}')

        for hashes, length, path in re.findall(r'(\d+)\t(\d+)\t(.+)\t.+\n', stdout):
            self.data[path] = (int(hashes), int(length))

    def _is_consistent(self):
        """Returns True if the sketch was generated from the genomes."""
        return set(self.data.keys()) == set(self.genomes.values())

    def _generate(self):
        """Generate a new sketch file."""
        with tempfile.TemporaryDirectory(prefix='gtdbtk_mash_tmp_') as dir_tmp:
            path_genomes = os.path.join(dir_tmp, 'genomes.txt')
            with open(path_genomes, 'w') as fh:
                for path in self.genomes.values():
                    fh.write(f'{path}\n')

            args = ['mash', 'sketch', '-l', '-p', self.cpus, path_genomes, '-o',
                    self.path, '-k', self.k, '-s', self.s]
            args = list(map(str, args))
            proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE, encoding='utf-8')
            with tqdm_log(total=len(self.genomes), unit='genome') as p_bar:
                for line in iter(proc.stderr.readline, ''):
                    if line.startswith('Sketching'):
                        p_bar.update()
            proc.wait()

            if proc.returncode != 0 or not os.path.isfile(self.path):
                raise GTDBTkExit(f'Error generating Mash sketch: {proc.stderr.read()}')


class QrySketchFile(SketchFile):
    name = 'user_query_sketch.msh'

    def __init__(self, genomes, root, prefix, cpus, k, s):
        """Create a query file for a given set of genomes.

        Parameters
        ----------
        genomes : dict[str, str]
            The genomes to create a sketch file from (genome_id, fasta_path).
        root : str
            The directory where the sketch file will be saved.
        prefix : str
            The prefix to use for this file.
        cpus : int
            The maximum number of CPUs available for Mash.
        k : int
            The k-mer size.
        s : int
            Maximum number of non-redundant hashes.
        """
        path = os.path.join(root, f'{prefix}.{self.name}')

        super().__init__(genomes, path, cpus, k, s)


class RefSketchFile(SketchFile):
    name = 'gtdb_ref_sketch.msh'

    def __init__(self, genomes, root, prefix, cpus, k, s, mash_db=None):
        """Create a query file for a given set of genomes.

        Parameters
        ----------
        genomes : dict[str, str]
            The genomes to create a sketch file from (genome_id, fasta_path).
        root : str
            The directory where the sketch file will be saved.
        prefix : str
            The prefix to use for this file.
        cpus : int
            The maximum number of CPUs available for Mash.
        k : int
            The k-mer size.
        s : int
            Maximum number of non-redundant hashes.
        mash_db : Optional[str]
            The path to read/write the pre-computed Mash reference sketch database.
        """
        if mash_db is not None:
            export_msh = mash_db.rstrip('\\')
            if not export_msh.endswith(".msh"):
                export_msh = export_msh + ".msh"
            if os.path.isdir(export_msh):
                raise GTDBTkExit(f"{export_msh} is a directory")
            make_sure_path_exists(os.path.dirname(export_msh))
            path = export_msh
        else:
            path = os.path.join(root, f'{prefix}.{self.name}')

        super().__init__(genomes, path, cpus, k, s)



def are_files_equal(path_left, path_right, ignore_order):
    """
    Determines if two files are equal based on their contents.

    :param path_left: The left file to compare.
    :param path_right: The right file to compare.
    :param ignore_order: True if the order of lines should be ignored, False otherwise.
    """

    with open(path_left, 'r') as f:
        lines_left = f.readlines()

    with open(path_right, 'r') as f:
        lines_right = f.readlines()

    if len(lines_left) != len(lines_right):
        return False

    if ignore_order:
        return set(lines_left).union(set(lines_right)) == set(lines_left)

    for line_left, line_right in zip(lines_left, lines_right):
        if line_left != line_right:
            return False
    return True



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
from typing import Dict

from gtdbtk.biolib_lite.common import make_sure_path_exists
from gtdbtk.config.output import PATH_AR122_PPLACER_CLASS, PATH_BAC120_PPLACER_CLASS
from gtdbtk.exceptions import GTDBTkExit


class PplacerClassifyFile(object):
    """Store the pplacer classifications."""

    def __init__(self, path: str):
        self.path: str = path
        self.data: Dict[str, str] = dict()

    def add_genome(self, gid: str, tax_str: str):
        """Adds the pplacer classification of a given genome."""
        if gid in self.data:
            raise GTDBTkExit(f'Warning! Attempting to add duplicate genome: {gid}')
        self.data[gid] = tax_str

    def write(self):
        """Write the file to disk."""
        make_sure_path_exists(os.path.dirname(self.path))
        with open(self.path, 'w') as fh:
            for gid, tax_str in sorted(self.data.items()):
                fh.write(f'{gid}\t{tax_str}\n')


class PplacerClassifyFileAR122(PplacerClassifyFile):
    """Store the pplacer classifications for the AR122 marker set."""

    def __init__(self, out_dir: str, prefix: str):
        path = os.path.join(out_dir, PATH_AR122_PPLACER_CLASS.format(prefix=prefix))
        super().__init__(path)


class PplacerClassifyFileBAC120(PplacerClassifyFile):
    """Store the pplacer classifications for the BAC120 marker set."""

    def __init__(self, out_dir: str, prefix: str):
        path = os.path.join(out_dir, PATH_BAC120_PPLACER_CLASS.format(prefix=prefix))
        super().__init__(path)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import random
from typing import Set

import dendropy

from gtdbtk.exceptions import GTDBTkExit


class RerootTree(object):
    """Reroot tree."""

    def __init__(self):
        """Initialization."""
        
        self.logger = logging.getLogger('timestamp')

    def root_with_outgroup(self, input_tree: str, output_tree: str, outgroup: Set[str]):
        """Reroot the tree using the given outgroup.

        Parameters
        ----------
        input_tree
          File containing Newick tree to rerooted.
        output_tree
          Name of file for rerooted tree.
        outgroup
          Labels of taxa in outgroup.
        """

        tree = dendropy.Tree.get_from_path(input_tree,
                                           schema='newick',
                                           rooting='force-rooted',
                                           preserve_underscores=True)

        outgroup = set(outgroup)
        outgroup_in_tree = set()
        ingroup_leaves = set()
        for n in tree.leaf_node_iter():
            if n.taxon.label in outgroup:
                outgroup_in_tree.add(n.taxon)
            else:
                ingroup_leaves.add(n)

        self.logger.info(f'Identified {len(outgroup_in_tree):,} outgroup taxa in the tree.')
        self.logger.info(f'Identified {len(ingroup_leaves):,} ingroup taxa in the tree.')

        if len(outgroup_in_tree) == 0:
            self.logger.error('No outgroup taxa identified in the tree.')
            raise GTDBTkExit('Tree was not rerooted.')

        # Since finding the MRCA is a rooted tree operation,
        # the tree is first rerooted on an ingroup taxa. This
        # ensures the MRCA of the outgroup can be identified
        # so long as the outgroup is monophyletic. If the
        # outgroup is polyphyletic trying to root on it
        # is ill defined. To try and pick a "good" root for
        # polyphyletic outgroups, random ingroup taxa are
        # selected until two of them give the same size
        # lineage. This will, likely, be the smallest
        # bipartition possible for the given outgroup though
        # this is not guaranteed.
        mrca = tree.mrca(taxa=outgroup_in_tree)
        mrca_leaves = len(mrca.leaf_nodes())
        while True:
            rnd_ingroup = random.sample(ingroup_leaves, 1)[0]
            tree.reroot_at_edge(rnd_ingroup.edge,
                                length1=0.5 * rnd_ingroup.edge_length,
                                length2=0.5 * rnd_ingroup.edge_length)

            mrca = tree.mrca(taxa=outgroup_in_tree)
            if len(mrca.leaf_nodes()) == mrca_leaves:
                break

            mrca_leaves = len(mrca.leaf_nodes())

        if len(mrca.leaf_nodes()) != len(outgroup_in_tree):
            self.logger.info('Outgroup is not monophyletic. Tree will be '
                             'rerooted at the MRCA of the outgroup.')
            self.logger.info(f'The outgroup consisted of '
                             f'{len(outgroup_in_tree):,} taxa, while the MRCA '
                             f'has {len(mrca.leaf_nodes()):,} leaf nodes.')
            if len(mrca.leaf_nodes()) == len(tree.leaf_nodes()):
                self.logger.warning('The MRCA spans all taxa in the tree.')
                self.logger.warning('This indicating the selected outgroup is '
                                    'likely polyphyletic in the current tree.')
                self.logger.warning('Polyphyletic outgroups are not suitable '
                                    'for rooting. Try another outgroup.')
        else:
            self.logger.info('Outgroup is monophyletic.')

        if mrca.edge_length is None:
            self.logger.info('Tree appears to already be rooted on this outgroup.')
        else:
            self.logger.info('Rerooting tree.')
            tree.reroot_at_edge(mrca.edge,
                                length1=0.5 * mrca.edge_length,
                                length2=0.5 * mrca.edge_length)
            tree.write_to_path(output_tree, schema='newick',
                               suppress_rooting=True, unquoted_underscores=True)
            self.logger.info(f'Rerooted tree written to: {output_tree}')

    def midpoint(self, input_tree, output_tree):
        """Reroot tree bat midpoint.

        Parameters
        ----------
        input_tree : str
          File containing Newick tree to rerooted.
        output_tree : str
          Name of file for rerooted tree.
        """

        tree = dendropy.Tree.get_from_path(
            input_tree, schema='newick', rooting='force-rooted', preserve_underscores=True)
        tree.reroot_at_midpoint()
        tree.write_to_path(output_tree, schema='newick',
                           suppress_rooting=True, unquoted_underscores=True)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import shutil
import tempfile
import unittest

from gtdbtk.biolib_lite.taxonomy import Taxonomy
from gtdbtk.exceptions import GTDBTkExit


class TestTaxonomy(unittest.TestCase):

    def setUp(self):
        self.dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')

    def tearDown(self):
        shutil.rmtree(self.dir_tmp)

    def test_read(self):
        expected = {
            'GCF_005435136.1': ['d__D', 'p__P', 'c__C', 'f__F', 'g__G', 's__S1'],
            '2': ['d__D', 'p__P', 'c__C', 'f__F', 'g__G', 's__S2']
        }
        path_tax = os.path.join(self.dir_tmp, 'tax_file.tsv')
        with open(path_tax, 'w') as f:
            for k, v in expected.items():
                f.write(f'{k}\t{";".join(v)}\n')
        t = Taxonomy()
        result = t.read(path_tax)
        self.assertDictEqual(expected, result)

    def test_read_canonical(self):
        to_write = {
            'GCF_005435136.1': ['d__D', 'p__P', 'c__C', 'f__F', 'g__G', 's__S1'],
            'RS_GCF_005435135.1': ['d__D', 'p__P', 'c__C', 'f__F', 'g__G', 's__S2']
        }
        expected = {
            'G005435136': to_write['GCF_005435136.1'],
            'G005435135': to_write['RS_GCF_005435135.1'],
        }
        path_tax = os.path.join(self.dir_tmp, 'tax_file.tsv')
        with open(path_tax, 'w') as f:
            for k, v in to_write.items():
                f.write(f'{k}\t{";".join(v)}\n')
        t = Taxonomy()
        result = t.read(path_tax, canonical_ids=True)
        self.assertDictEqual(expected, result)

    def test_read_error(self):
        expected = {
            'GCF_005435136.1': ['d__D', 'p__P', 'c__C', 'f__F', 'g__G', 's__S']
        }
        path_tax = os.path.join(self.dir_tmp, 'tax_file.tsv')
        with open(path_tax, 'w') as f:
            for k, v in expected.items():
                f.write(f'{k},{";".join(v)}\n')
        t = Taxonomy()
        self.assertRaises(GTDBTkExit, t.read, path_tax)



import argparse
import tempfile
from contextlib import contextmanager

from gtdbtk.biolib_lite.custom_help_formatter import ChangeTempAction
from gtdbtk.biolib_lite.custom_help_formatter import CustomHelpFormatter
from gtdbtk.config.config import AF_THRESHOLD


@contextmanager
def subparser(parser, name, desc):
    yield parser.add_parser(name, conflict_handler='resolve', help=desc,
                            formatter_class=CustomHelpFormatter)


@contextmanager
def mutex_group(parser, required):
    group = parser.add_argument_group(f'mutually exclusive '
                                      f'{"required" if required else "optional"} '
                                      f'arguments')
    yield group.add_mutually_exclusive_group(required=required)


@contextmanager
def arg_group(parser, name):
    yield parser.add_argument_group(name)


def __temp_dir(group):
    group.add_argument('--tmpdir', action=ChangeTempAction, default=tempfile.gettempdir(),
                       help="specify alternative directory for temporary files")


def __genome_dir(group):
    group.add_argument('--genome_dir', help="directory containing genome files in FASTA format")


def __batchfile(group):
    group.add_argument('--batchfile', help="path to file describing genomes - tab "
                                           "separated in 2 or 3 columns (FASTA "
                                           "file, genome ID, translation "
                                           "table [optional])")


def __archaea(group):
    group.add_argument('--archaea', action='store_true', default=False,
                       help='process archaeal genomes')


def __bacteria(group):
    group.add_argument('--bacteria', action='store_true', default=False,
                       help='process bacterial genomes')


def __outgroup_taxon(group, required):
    group.add_argument('--outgroup_taxon', type=str, default=None, required=required,
                       help="taxon to use as outgroup (e.g., "
                            "``p__Patescibacteria`` or ``p__Altarchaeota``)")


def __out_dir(group, required):
    group.add_argument('--out_dir', type=str, default=None, required=required,
                       help="directory to output files")


def __extension(group):
    group.add_argument('-x', '--extension', type=str, default='fna',
                       help='extension of files to process, ``gz`` = gzipped')


def __skip_gtdb_refs(group):
    group.add_argument('--skip_gtdb_refs', action="store_true", default=False,
                       help='do not include GTDB reference genomes in multiple sequence alignment')


def __taxa_filter(group):
    group.add_argument('--taxa_filter', type=str, default=None,
                       help=('filter GTDB genomes to taxa (comma separated) within '
                             + 'specific taxonomic groups (e.g.: ``d__Bacteria`` '
                             + 'or ``p__Proteobacteria,p__Actinobacteria``)'))


def __min_perc_aa(group):
    group.add_argument('--min_perc_aa', type=float, default=10,
                       help='exclude genomes that do not have at least this '
                            'percentage of AA in the MSA (inclusive bound)')


def __custom_msa_filters(group):
    group.add_argument('--custom_msa_filters', action="store_true", default=False,
                       help=('perform custom filtering of MSA with ``cols_per_gene``, ``min_consensus`` '
                             + '``max_consensus``, and ``min_perc_taxa`` parameters instead of using canonical mask'))


def __cols_per_gene(group):
    group.add_argument('--cols_per_gene', type=int, default=42,
                       help='maximum number of columns to retain per gene '
                            'when generating the MSA')


def __min_consensus(group):
    group.add_argument('--min_consensus', type=float, default=25,
                       help='minimum percentage of the same amino acid '
                            'required to retain column (inclusive bound)')


def __max_consensus(group):
    group.add_argument('--max_consensus', type=float, default=95,
                       help='maximum percentage of the same amino acid '
                            'required to retain column (exclusive bound)')


def __min_perc_taxa(group):
    group.add_argument('--min_perc_taxa', type=float, default=50,
                       help='minimum percentage of taxa required to retain column (inclusive bound)')


def __prot_model(group):
    group.add_argument('--prot_model', choices=['JTT', 'WAG', 'LG'],
                       help='protein substitution model for tree inference', default='WAG')


def __rnd_seed(group):
    group.add_argument('--rnd_seed', type=int, default=None,
                       help='random seed to use for selecting columns, e.g. ``42``')


def __no_support(group):
    group.add_argument('--no_support', action="store_true", default=False,
                       help="do not compute local support values using the Shimodaira-Hasegawa test")


def __gamma(group):
    group.add_argument('--gamma', action="store_true", default=False,
                       help="rescale branch lengths to optimize the Gamma20 likelihood")


def __gtdbtk_classification_file(group):
    group.add_argument('--gtdbtk_classification_file', type=str, default=None,
                       help="file with GTDB-Tk classifications produced by the `classify` command")


def __custom_taxonomy_file(group):
    group.add_argument('--custom_taxonomy_file', type=str, default=None,
                       help="file indicating custom taxonomy strings for user "
                            "genomes, that should contain any genomes belonging to the outgroup. "
                            "Format: GENOME_ID<TAB>d__;p__;c__;o__;f__;g__;s__")


def __prefix(group):
    group.add_argument('--prefix', type=str, default='gtdbtk',
                       help='prefix for all output files')


def __cpus(group):
    group.add_argument('--cpus', default=1, type=int, help='number of CPUs to use')


def __force(group):
    group.add_argument('--force', action='store_true', default=False,
                       help='continue processing if an error occurs on a single genome')


def __debug(group):
    group.add_argument('--debug', action="store_true", default=False,
                       help='create intermediate files for debugging purposes')


def __help(group):
    group.add_argument('-h', '--help', action="help", help="show help message")


def __pplacer_cpus(group):
    group.add_argument('--pplacer_cpus', type=int, default=None,
                       help='use ``pplacer_cpus`` during placement (default: ``cpus``)')


def __scratch_dir(group):
    group.add_argument('--scratch_dir', type=str, default=None,
                       help='Reduce pplacer memory usage by writing to disk (slower).')


def __recalculate_red(group):
    group.add_argument('-r', '--recalculate_red', default=False, action='store_true',
                       help='recalculate RED values based on the reference tree and all added user genomes')


def __split_tree(group):
    group.add_argument('-s', '--split_tree', default=False, action='store_true',
                       help='Use shards of the reference tree (for Bacteria only). reduce memory usage (slower).')


def __identify_dir(group, required):
    group.add_argument('--identify_dir', type=str, default=None, required=required,
                       help="output directory of 'identify' command")


def __skip_trimming(group):
    group.add_argument('--skip_trimming', default=False, action="store_true",
                       help='skip the trimming step and return the full MSAs')


def __msa_file(group, required):
    group.add_argument('--msa_file', type=str, default=None, required=required,
                       help="multiple sequence alignment in FASTA format")


def __align_dir(group, required):
    group.add_argument('--align_dir', type=str, default=None, required=required,
                       help="output directory of 'align' command")


def __input_tree(group, required):
    group.add_argument('--input_tree', type=str, default=None, required=required,
                       help="path to the unrooted tree in Newick format")


def __input_tree__rooted(group, required):
    group.add_argument('--input_tree', type=str, default=None, required=required,
                       help="rooted input tree with labelled ingroup taxon")


def __output_tree(group, required):
    group.add_argument('--output_tree', type=str, default=None, required=required,
                       help='path to output the tree')


def __ingroup_taxon(group, required):
    group.add_argument('--ingroup_taxon', type=str, default=None, required=required,
                       help="labelled ingroup taxon to use as root for "
                            "establishing RED values (e.g., c__Bacilli or f__Lactobacillaceae")


def __no_mash(group):
    group.add_argument('--no_mash', default=False, action='store_true',
                       help='skip pre-filtering of genomes using Mash')


def __mash_k(group):
    group.add_argument('--mash_k', default=16, type=int, help='k-mer size [1-32]')


def __mash_s(group):
    group.add_argument('--mash_s', default=5000, type=int, help='maximum number of non-redundant hashes')


def __mash_d(group):
    group.add_argument('--mash_d', default=0.1, type=float, help='maximum distance to keep [0-1]')


def __mash_v(group):
    group.add_argument('--mash_v', default=1.0, type=float, help='maximum p-value to keep [0-1]')


def __mash_db(group):
    group.add_argument('--mash_db', default=None, type=str,
                       help='path to save/read (if exists) the Mash reference sketch database (.msh)')


def __min_af(group):
    group.add_argument('--min_af', type=float, default=AF_THRESHOLD,
                       help='minimum alignment fraction to consider closest genome')


def __untrimmed_msa(group, required):
    group.add_argument('--untrimmed_msa', type=str, default=None, required=required,
                       help="path to the untrimmed MSA file")


def __output(group, required):
    group.add_argument('--output', type=str, default=None, required=required,
                       help='output file')


def __mask_file(group):
    group.add_argument('--mask_file', type=str, default=None,
                       help="path to a custom mask file for trimming the MSA")


def __reference_mask(group):
    group.add_argument('--reference_mask',
                       choices=['arc', 'bac'],
                       help="reference mask already present in GTDB-Tk")


def __domain(group, required):
    group.add_argument('--domain', required=required, choices=['arc', 'bac'],
                       help="domain to export")


def __write_single_copy_genes(group):
    group.add_argument('--write_single_copy_genes', default=False, action='store_true',
                       help='output unaligned single-copy marker genes')


def get_main_parser():
    # Setup the main, and sub parsers.
    main_parser = argparse.ArgumentParser(prog='gtdbtk', add_help=False, conflict_handler='resolve')
    sub_parsers = main_parser.add_subparsers(help="--", dest='subparser_name')

    # de novo workflow.
    with subparser(sub_parsers, 'de_novo_wf', 'Infer de novo tree and decorate with GTDB taxonomy.') as parser:
        with mutex_group(parser, required=True) as grp:
            __genome_dir(grp)
            __batchfile(grp)
        with mutex_group(parser, required=True) as grp:
            __bacteria(grp)
            __archaea(grp)
        with arg_group(parser, 'required named arguments') as grp:
            __outgroup_taxon(grp, required=True)
            __out_dir(grp, required=True)
        with arg_group(parser, 'optional arguments') as grp:
            __extension(grp)
            __skip_gtdb_refs(grp)
            __taxa_filter(grp)
            __min_perc_aa(grp)
            __custom_msa_filters(grp)
            __cols_per_gene(grp)
            __min_consensus(grp)
            __max_consensus(grp)
            __min_perc_taxa(grp)
            __rnd_seed(grp)
            __prot_model(grp)
            __no_support(grp)
            __gamma(grp)
            __gtdbtk_classification_file(grp)
            __custom_taxonomy_file(grp)
            __prefix(grp)
            __cpus(grp)
            __force(grp)
            __temp_dir(grp)
            __debug(grp)
            __help(grp)

    # Classify workflow.
    with subparser(sub_parsers, 'classify_wf', 'Classify genomes by placement in GTDB reference tree.') as parser:
        with mutex_group(parser, required=True) as grp:
            __genome_dir(grp)
            __batchfile(grp)
        with arg_group(parser, 'required named arguments') as grp:
            __out_dir(grp, required=True)
        with arg_group(parser, 'optional arguments') as grp:
            __extension(grp)
            __min_perc_aa(grp)
            __prefix(grp)
            __cpus(grp)
            __pplacer_cpus(grp)
            __force(grp)
            __scratch_dir(grp)
            # __recalculate_red(grp)
            # __split_tree(grp)
            __min_af(grp)
            __temp_dir(grp)
            __debug(grp)
            __help(grp)

    # Identify marker genes in genomes.
    with subparser(sub_parsers, 'identify', 'Identify marker genes in genomes.') as parser:
        with mutex_group(parser, required=True) as grp:
            __genome_dir(grp)
            __batchfile(grp)
        with arg_group(parser, 'required named arguments') as grp:
            __out_dir(grp, required=True)
        with arg_group(parser, 'optional arguments') as grp:
            __extension(grp)
            __prefix(grp)
            __cpus(grp)
            __force(grp)
            __write_single_copy_genes(grp)
            __temp_dir(grp)
            __debug(grp)
            __help(grp)

    # Create a multiple sequence alignment.
    with subparser(sub_parsers, 'align', 'Create multiple sequence alignment.') as parser:
        with arg_group(parser, 'required named arguments') as grp:
            __identify_dir(grp, required=True)
            __out_dir(grp, required=True)
        with arg_group(parser, 'optional arguments') as grp:
            __skip_gtdb_refs(grp)
            __taxa_filter(grp)
            __min_perc_aa(grp)
            __cols_per_gene(grp)
            __min_consensus(grp)
            __max_consensus(grp)
            __min_perc_taxa(grp)
            __rnd_seed(grp)
            __prefix(grp)
            __cpus(grp)
            __temp_dir(grp)
            __debug(grp)
            __help(grp)
        with mutex_group(parser, required=False) as grp:
            __custom_msa_filters(grp)
            __skip_trimming(grp)

    # Infer a de novo tree.
    with subparser(sub_parsers, 'infer', 'Infer tree from multiple sequence alignment.') as parser:
        with arg_group(parser, 'required named arguments') as grp:
            __msa_file(grp, required=True)
            __out_dir(grp, required=True)
        with arg_group(parser, 'optional arguments') as grp:
            __prot_model(grp)
            __no_support(grp)
            __gamma(grp)
            __prefix(grp)
            __cpus(grp)
            __temp_dir(grp)
            __debug(grp)
            __help(grp)

    # Classify genomes via placement with pplacer.
    with subparser(sub_parsers, 'classify', 'Determine taxonomic classification of genomes.') as parser:
        with mutex_group(parser, required=True) as grp:
            __genome_dir(grp)
            __batchfile(grp)
        with arg_group(parser, 'required named arguments') as grp:
            __align_dir(grp, required=True)
            __out_dir(grp, required=True)
        with arg_group(parser, 'optional arguments') as grp:
            __extension(grp)
            __prefix(grp)
            __cpus(grp)
            __pplacer_cpus(grp)
            __scratch_dir(grp)
            # __split_tree(grp)
            # __recalculate_red(grp)
            __min_af(grp)
            __temp_dir(grp)
            __debug(grp)
            __help(grp)

    # Root a tree using an outgroup.
    with subparser(sub_parsers, 'root', 'Root tree using an outgroup.') as parser:
        with arg_group(parser, 'required named arguments') as grp:
            __input_tree(grp, required=True)
            __outgroup_taxon(grp, required=True)
            __output_tree(grp, required=True)
        with arg_group(parser, 'optional arguments') as grp:
            __gtdbtk_classification_file(grp)
            __custom_taxonomy_file(grp)
            __temp_dir(grp)
            __debug(grp)
            __help(grp)

    # Decorate a tree.
    with subparser(sub_parsers, 'decorate', 'Decorate tree with GTDB taxonomy') as parser:
        with arg_group(parser, 'required named arguments') as grp:
            __input_tree(grp, required=True)
            __output_tree(grp, required=True)
        with arg_group(parser, 'optional arguments') as grp:
            __gtdbtk_classification_file(grp)
            __custom_taxonomy_file(grp)
            __temp_dir(grp)
            __debug(grp)
            __help(grp)

    # Establish taxonomic ranks of internal nodes using RED
    with subparser(sub_parsers, 'infer_ranks', 'Establish taxonomic ranks of internal nodes using RED.') as parser:
        with arg_group(parser, 'required named arguments') as grp:
            __input_tree__rooted(grp, required=True)
            __ingroup_taxon(grp, required=True)
            __output_tree(grp, required=True)
        with arg_group(parser, 'optional arguments') as grp:
            __temp_dir(grp)
            __debug(grp)
            __help(grp)

    # Calculate ANI to representative genomes.
    with subparser(sub_parsers, 'ani_rep', 'Calculate ANI to GTDB representative genomes.') as parser:
        with mutex_group(parser, required=True) as grp:
            __genome_dir(grp)
            __batchfile(grp)
        with arg_group(parser, 'required named arguments') as grp:
            __out_dir(grp, required=True)
        with arg_group(parser, 'optional Mash arguments') as grp:
            __no_mash(grp)
            __mash_k(grp)
            __mash_s(grp)
            __mash_d(grp)
            __mash_v(grp)
            __mash_db(grp)
        with arg_group(parser, 'optional FastANI arguments') as grp:
            __min_af(grp)
        with arg_group(parser, 'optional arguments') as grp:
            __extension(grp)
            __prefix(grp)
            __cpus(grp)
            __temp_dir(grp)
            __debug(grp)
            __help(grp)

    # Run a test.
    with subparser(sub_parsers, 'test', 'Test the classify_wf pipeline with 3 archaeal genomes.') as parser:
        with arg_group(parser, 'optional arguments') as grp:
            __out_dir(grp, required=False)
            __cpus(grp)
            __debug(grp)
            __help(grp)

    # Trim MSA.
    with subparser(sub_parsers, 'trim_msa', 'Trim an untrimmed MSA file based on a mask.') as parser:
        with arg_group(parser, 'required named arguments') as grp:
            __untrimmed_msa(grp, required=True)
            __output(grp, required=True)
        with mutex_group(parser, required=True) as grp:
            __mask_file(grp)
            __reference_mask(grp)
        with arg_group(parser, 'optional arguments') as grp:
            __debug(grp)
            __help(grp)

    # Export MSA.
    with subparser(sub_parsers, 'export_msa', 'Export the untrimmed archaeal or bacterial MSA file.') as parser:
        with arg_group(parser, 'required named arguments') as grp:
            __domain(grp, required=True)
            __output(grp, required=True)
        with arg_group(parser, 'optional arguments') as grp:
            __debug(grp)
            __help(grp)

    # Verify install.
    with subparser(sub_parsers, 'check_install', 'Verify third party programs and '
                                                 'GTDB reference package.') as parser:
        with arg_group(parser, 'optional arguments') as grp:
            __debug(grp)
            __help(grp)

    return main_parser



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import shutil
import tempfile
import unittest
from collections import defaultdict

from gtdbtk.biolib_lite.seq_io import write_fasta, read_fasta


class TestBiolibLiteCommon(unittest.TestCase):

    def setUp(self):
        self.dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')

    def tearDown(self):
        shutil.rmtree(self.dir_tmp)

    def test_write_fasta(self):
        """ Test that the fasta file is written correctly with default parameters """
        path_fasta = os.path.join(self.dir_tmp, 'fasta.fna')
        seqs = {'genome_1': 'CAGTTCAGTT', 'genome_2': 'TTAGTCA', 'genome_3': 'CAG'}
        write_fasta(seqs, path_fasta)

    def test_write_fasta_wrap(self):
        """ Test that the sequences are wrapped as specified """
        path_fasta = os.path.join(self.dir_tmp, 'fasta.fna')
        seqs = {'genome_1': 'CAGTTCAGTT', 'genome_2': 'TTAGTCA', 'genome_3': 'CAG'}
        write_fasta(seqs, path_fasta, wrap=4)

        self.assertDictEqual(seqs, read_fasta(path_fasta))

        file_content = defaultdict(list)
        with open(path_fasta, 'r') as f:
            cur_gid = None
            for line in f.readlines():
                line = line.strip()
                if line.startswith('>'):
                    cur_gid = line[1:]
                else:
                    file_content[cur_gid].append(line)

        for gid, seq_list in file_content.items():
            for cur_seq in seq_list:
                self.assertLessEqual(len(cur_seq), 4)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################
import os
import shutil
import tempfile
import unittest
import json
from gtdbtk.external.fastani import FastANI
import gtdbtk.config.config as Config

class TestFastANI(unittest.TestCase):

    def setUp(self):
        self.dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        self.cpus = 1
        self.genome_root = Config.FASTANI_GENOMES

    def tearDown(self):
        shutil.rmtree(self.dir_tmp)

    def test_run(self):
        """Test that FastANI produces the expected output (version dependent)"""

        fa = FastANI(self.cpus, force_single=True)

        """
        a = d__Archaea; p__Altiarchaeota; c__Altiarchaeia; o__Altiarchaeales; f__Altiarchaeaceae; g__Altiarchaeum; s__Altiarchaeum sp001873845
        b = d__Archaea; p__Altiarchaeota; c__Altiarchaeia; o__Altiarchaeales; f__Altiarchaeaceae; g__Altiarchaeum; s__Altiarchaeum sp002083985
        c = d__Bacteria; p__Aquificota; c__Desulfurobacteriia; o__Desulfurobacteriales; f__Desulfurobacteriaceae; g__Thermovibrio; s__Thermovibrio ammonificans
        x = d__Archaea; p__Altiarchaeota; c__Altiarchaeia; o__GCA-002841105; f__GCA-002841105; g__GCA-002841105; s__GCA-002841105 sp002841105
        y = d__Bacteria; p__Aerophobota; c__Aerophobia; o__Aerophobales; f__Aerophobaceae; g__Aerophobus; s__Aerophobus profundus
        z = d__Archaea; p__Halobacterota; c__Archaeoglobi; o__JdFR-21; f__JdFR-21; g__JdFR-21; s__JdFR-21 sp002011165
        """
        d_compare = {'a': {'x', 'y'},
                     'b': {'x'},
                     'c': {'z'}}
        d_paths = {'a': os.path.join(self.genome_root,'GCA/001/873/845', 'GCA_001873845.1_genomic.fna.gz'),
                   'b': os.path.join(self.genome_root,'GCA/002/083/985', 'GCA_002083985.1_genomic.fna.gz'),
                   'c': os.path.join(self.genome_root,'GCF/000/185/805', 'GCF_000185805.1_genomic.fna.gz'),
                   'x': os.path.join(self.genome_root,'GCA/002/841/105', 'GCA_002841105.1_genomic.fna.gz'),
                   'y': os.path.join(self.genome_root,'GCA/000/402/295', 'GCA_000402295.1_genomic.fna.gz'),
                   'z': os.path.join(self.genome_root,'GCA/002/011/165', 'GCA_002011165.1_genomic.fna.gz')}

        result = fa.run(d_compare, d_paths)

        expected = {'a': {'x': {'ani': 82.5201, 'af': 0.57},
                          'y': {'ani': 74.5154, 'af': 0.0}},
                    'b': {'x': {'ani': 84.5846, 'af': 0.55}},
                    'c': {'z': {'ani': 74.4978, 'af': 0.01}}}
        self.assertEqual(json.dumps(result, sort_keys=True), json.dumps(expected, sort_keys=True))

    def test_parse_output_file(self):
        fa = FastANI(self.cpus, force_single=True)

        out_txt = [['q1', 'r1', '83.1234', '5', '10'],
                   ['q1', 'r2', '99.1111', '3', '10'],
                   ['q2', 'r1', '1.12', '555', '1111']]

        expected = {'q1': {'r1': (83.1234, 0.5), 'r2': (99.1111, 0.3)}, 'q2': {'r1': (1.12, 0.5)}}

        path_f1 = os.path.join(self.dir_tmp, 'f1.txt')
        path_f2 = os.path.join(self.dir_tmp, 'f2.txt')
        path_f3 = os.path.join(self.dir_tmp, 'f3.txt')

        with open(path_f1, 'w') as fh:
            for x in out_txt:
                fh.write(' '.join(x) + '\n')

        result_f1 = fa.parse_output_file(path_f1)
        self.assertEqual(result_f1, expected)

        with open(path_f2, 'w') as fh:
            for x in out_txt:
                fh.write('\t'.join(x) + '\n')

        result_f2 = fa.parse_output_file(path_f2)
        self.assertEqual(result_f2, expected)

        open(path_f3, 'w').close()
        result_f3 = fa.parse_output_file(path_f3)
        self.assertEqual(result_f3, {})



import logging

from gtdbtk.exceptions import GTDBTkExit


class Batchfile(object):

    def __init__(self, path: str):
        self.path = path
        self.genome_path, self.genome_tln = self.read(path)

    @staticmethod
    def read(path):
        logger = logging.getLogger('timestamp')
        genomes, tln_tables = dict(), dict()
        seen_paths = set()
        warnings = list()
        with open(path) as fh:
            for line_no, line in enumerate(fh):
                line_split = line.strip().split("\t")
                if line_split[0] == '':
                    continue  # blank line

                if len(line_split) not in {2, 3}:
                    raise GTDBTkExit('Batch file must contain either 2 '
                                     'columns (detect translation table), '
                                     'or 3 (specify translation table).')

                if len(line_split) == 2:
                    genome_file, genome_id = line_split
                elif len(line_split) == 3:
                    genome_file, genome_id, tln_table = line_split
                    if tln_table not in {'4', '11'}:
                        raise GTDBTkExit('Specified translation table must '
                                         'be either 4, or 11.')
                    tln_tables[genome_id] = int(tln_table)

                if genome_file is None or genome_file == '':
                    warnings.append(f'Missing genome path on line {line_no + 1}.')
                elif genome_id is None or genome_id == '':
                    warnings.append(f'Missing genome ID on line {line_no + 1}.')
                elif genome_id in genomes:
                    warnings.append(f'Genome ID {genome_id} appears multiple times.')
                if genome_file in seen_paths:
                    logger.warning(f'Genome file appears multiple times: {genome_file}')

                # All good, record the value.
                genomes[genome_id] = genome_file
                seen_paths.add(genome_file)

        # Check if any warnings were raised.
        if len(warnings) > 0:
            warning_str = '\n'.join(warnings)
            raise GTDBTkExit(f'Please check the format of your batchfile, '
                             f'the following errors were found: {warning_str}')

        return genomes, tln_tables
